# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx
import numpy as np

import sockeye.constants as C
import sockeye.transformer


def test_auto_regressive_bias_op():
    bias = mx.nd.Custom(op_type='auto_regressive_bias', length=2)

    assert bias.dtype == np.float32

    expected = np.array([[0.0, -1.0e8], [0.0, 0.0]]).reshape((1, 2, 2))
    np.testing.assert_array_equal(bias.asnumpy(), expected)


def test_auto_regressive_bias_op_float16():
    bias = mx.nd.Custom(op_type='auto_regressive_bias', length=2, dtype=C.DTYPE_FP16)

    assert bias.dtype == np.float16

    expected = np.array([[0.0, -49152.0], [0.0, 0.0]]).reshape((1, 2, 2))
    np.testing.assert_array_equal(bias.asnumpy(), expected)


def test_auto_regressive_bias_sym():
    bias = mx.sym.Custom(op_type='auto_regressive_bias', length=2)

    arg_types, out_types, aux_types = bias.infer_type()
    assert out_types[0] == np.float32

    out = bias.eval()[0]

    assert out.dtype == np.float32

    expected = np.array([[0.0, -1.0e8], [0.0, 0.0]]).reshape((1, 2, 2))
    np.testing.assert_array_equal(out.asnumpy(), expected)


def test_auto_regressive_bias_sym_float16():
    bias = mx.sym.Custom(op_type='auto_regressive_bias', length=2, dtype=C.DTYPE_FP16)

    arg_types, out_types, aux_types = bias.infer_type()
    assert out_types[0] == np.float16

    out = bias.eval()[0]

    assert out.dtype == np.float16

    expected = np.array([[0.0, -49152.0], [0.0, 0.0]]).reshape((1, 2, 2))
    np.testing.assert_array_equal(out.asnumpy(), expected)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
A set of utility methods for images.
"""
import os
from shutil import copyfile
from typing import List, Optional

import numpy as np

from ..log import setup_main_logger

# Temporary logger, the real one (logging to a file probably, will be created
# in the main function)
logger = setup_main_logger(__name__, file_logging=False, console=True)

try:  # Try to import pillow
    from PIL import Image  # pylint: disable=import-error
except ImportError as e:
    raise RuntimeError("Please install pillow.")

def copy_mx_model_to(model_path, model_epoch, output_folder):
    """
    Copy mxnet models to new path

    :param model_path: Model path without -symbol.json and -%04d.params
    :param model_epoch: Epoch of the pretrained model
    :param output_folder: Output folder
    :return: New folder where the files are moved to
    """
    target_path = os.path.join(output_folder, os.path.basename(model_path))
    logger.info("Copying image model from {} to {}".format(model_path,
                                                           target_path))
    suffix = ['-symbol.json', '-%04d.params' % (model_epoch,)]
    for s in suffix:
        copyfile(model_path+s, target_path+s)
    return target_path


def crop_resize_image(image, size):
    """
    Resize the input image.

    :param image: Original image which is a  PIL object.
    :param size: Tuple of height and width to resize the image to.
    :return: Resized image which is a PIL object
    """
    width, height = image.size
    if width > height:
        left = (width - height) / 2
        right = width - left
        top = 0
        bottom = height
    else:
        top = (height - width) / 2
        bottom = height - top
        left = 0
        right = width
    image = image.crop((left, top, right, bottom))
    image = image.resize(size, Image.ANTIALIAS)
    return image


def load_preprocess_images(image_paths: List[str], image_size: tuple) -> List:
    """
    Load and pre-process the images specified with absolute paths.

    :param image_paths: List of images specified with paths.
    :param image_size: Tuple to resize the image to (Channels, Height, Width)
    :return: A list of loaded images (numpy arrays).
    """
    image_size = image_size[1:]  # we do not need the number of channels
    images = []
    for image_path in image_paths:
        images.append(load_preprocess_image(image_path, image_size))
    return images


def load_preprocess_image(image_path: str, image_size: tuple):
    with Image.open(image_path) as image:
        image_o = preprocess_image(image, image_size)
    return image_o


def preprocess_image(image: Image, image_size: tuple):
    # Resize to fixed input
    image_o = crop_resize_image(image, image_size)
    # convert to numpy
    image_o = np.asarray(image_o)
    # Gray-level to 3 channels
    if len(image_o.shape)==2:
        image_o = np.tile(image_o[:,:,None], (1,1,3))
    # (height, width, channel) -> (channel, height, width)
    image_o = np.swapaxes(image_o, 0, 2)
    image_o = np.swapaxes(image_o, 1, 2)
    return image_o


def load_features(paths: List[str],
                  expected_shape: Optional[tuple] = None) -> List:
    """
    Load features specified with absolute paths.

    :param paths: List of files specified with paths.
    :return: A list of loaded images (numpy arrays).
    """
    data = []
    for path in paths:
        data.append(load_feature(path, expected_shape))
    return data


def load_feature(path: str,
                 expected_shape: Optional[tuple] = None) -> np.ndarray:
    try: # compresse
        data = np.load(path)['data']
    except IndexError:  # uncompressed
        data = np.load(path)
    if expected_shape is not None:
        np.testing.assert_array_equal(data.shape, expected_shape,
                  err_msg="Loaded feature shape different than provided one. "
                          "(current: {}, provided{})".format(data.shape,
                                                             expected_shape))
    return data


def save_features(paths: List[str], datas: List[np.ndarray],
                  compressed:bool = False) -> List:
    """
    Save features specified with absolute paths.

    :param paths: List of files specified with paths.
    :param datas: List of numpy ndarrays to save into the respective files
    :param compressed: Use numpy compression
    :return: A list of file names.
    """
    fnames = []
    for path, data in zip(paths, datas):
        fnames.append(save_feature(path, data, compressed))
    return fnames


def save_feature(path: str,
                 data: np.ndarray,
                 compressed:bool = False) -> None:
    if compressed:
        np.savez_compressed(path, data=data)
        path += ".npz"
    else:
        np.save(path, data)
        path += ".npy"
    return path


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import random
import string

import pytest

from test.common_image_captioning import run_train_captioning, tmp_img_captioning_dataset

_LINE_MAX_LENGTH = 9
_TEST_MAX_LENGTH = 20

ENCODER_DECODER_SETTINGS = [
    # 2-layer LSTM decoder with attention
    ("--encoder image-pretrain-cnn --image-encoder-num-hidden 8 --decoder rnn --rnn-cell-type lstm "
     "--batch-type sentence --batch-size 2 "
     "--initial-learning-rate 0.0003 --gradient-clipping-threshold 1.0 --bucket-width 2 "
     "--rnn-num-hidden 8 --rnn-decoder-state-init zero --weight-normalization "
     "--checkpoint-frequency 2 --max-updates 2 --num-layers 1:2 ",
     "--beam-size 2"),
    # LSTM decoder with attention: no global, encoder hiddens 8, rnn last, load all feats to mem
    ("--encoder image-pretrain-cnn --image-encoder-num-hidden 8 --no-image-encoder-global-descriptor "
     "--decoder rnn --rnn-cell-type lstm --batch-size 12 --optimizer adam --load-all-features-to-memory "
     "--initial-learning-rate 0.0003 --gradient-clipping-threshold 1.0 --bucket-width 2 "
     "--rnn-num-hidden 8 --rnn-decoder-state-init last --weight-normalization "
     "--checkpoint-frequency 2 --max-updates 2",
     "--beam-size 2"),
    # Transformer decoder
    ("--encoder image-pretrain-cnn --image-encoder-num-hidden 8 --decoder transformer --batch-size 12 --num-embed 4 "
     "--transformer-attention-heads 2 --transformer-model-size 4 --transformer-feed-forward-num-hidden 8 "
     "--initial-learning-rate 0.0003 --gradient-clipping-threshold 1.0 --bucket-width 2 "
     "--checkpoint-frequency 2 --max-updates 2",
     "--beam-size 2"),
    # 2-layer CNN decoder
    ("--encoder image-pretrain-cnn --decoder cnn --num-layers 2 --batch-size 12 "
     "--initial-learning-rate 0.0003 "
     "--cnn-num-hidden 8 --image-encoder-num-hidden 8 --cnn-positional-embedding-type fixed "
     "--checkpoint-frequency 2 --max-updates 2",
     "--beam-size 2")
]


@pytest.mark.parametrize("train_params, translate_params",
                         ENCODER_DECODER_SETTINGS)
def test_caption_random_features(train_params: str, translate_params: str):
    # generate random names
    source_list = [''.join(random.choice(string.ascii_uppercase) for _ in range(4)) for i in range(15)]
    prefix = "tmp_caption_ramdom"
    use_features = True
    with tmp_img_captioning_dataset(source_list,
                                    prefix,
                                    train_max_length=_LINE_MAX_LENGTH,
                                    dev_max_length=_LINE_MAX_LENGTH,
                                    test_max_length=_TEST_MAX_LENGTH,
                                    use_features=use_features) as data:
        # Test model configuration, including the output equivalence of batch and no-batch decoding
        translate_params_batch = translate_params + " --batch-size 2"

        # Ignore return values (perplexity and BLEU) for integration test
        run_train_captioning(train_params=train_params,
                             translate_params=translate_params,
                             translate_params_equiv=translate_params_batch,
                             train_source_path=data['source'],
                             train_target_path=data['target'],
                             dev_source_path=data['validation_source'],
                             dev_target_path=data['validation_target'],
                             test_source_path=data['test_source'],
                             test_target_path=data['test_target'],
                             max_seq_len=_LINE_MAX_LENGTH + 1,
                             work_dir=data['work_dir'])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from tempfile import TemporaryDirectory

import mxnet as mx

from sockeye import constants as C
from sockeye.image_captioning.encoder import ImageLoadedCnnEncoderConfig, \
    ImageLoadedCnnEncoder
from test.common_image_captioning import create_simple_and_save_to_disk


def test_image_loaded_cnn_encoder():
    layer_name = "conv2"
    encoded_seq_len = 16 + 1  # +1 for global descriptor
    num_embed = 10
    no_global_descriptor = False
    preextracted_features = False
    source_image_size = (3, 20, 20)
    batch_size = 8

    with TemporaryDirectory() as work_dir:
        model_path = os.path.join(work_dir, '2-conv-layer')
        epoch = 0
        # Create net and save to disk
        create_simple_and_save_to_disk(model_path, epoch, source_image_size, batch_size)
        # Setup encoder
        image_cnn_encoder_config = ImageLoadedCnnEncoderConfig(
                                        model_path=model_path,
                                        epoch=epoch,
                                        layer_name=layer_name,
                                        encoded_seq_len=encoded_seq_len,
                                        num_embed=num_embed,
                                        no_global_descriptor=no_global_descriptor,
                                        preextracted_features=preextracted_features)
        image_cnn_encoder = ImageLoadedCnnEncoder(image_cnn_encoder_config)
        # Prepare for inference
        data_nd = mx.nd.random_normal(shape=(batch_size,) + source_image_size)
        source = mx.sym.Variable(C.SOURCE_NAME)
        embedding, encoded_data_length, seq_len = image_cnn_encoder.encode(source,
                                                                           None,
                                                                           None)
        data_names = ['source']
        module = mx.mod.Module(symbol=embedding,
                               data_names=data_names,
                               label_names=None)
        module.bind(for_training=False,
                    data_shapes=[(data_names[0], (batch_size,) + source_image_size)])
        # Pretrained net
        initializers = image_cnn_encoder.get_initializers()
        init = mx.initializer.Mixed(*zip(*initializers))
        module.init_params(init)
        provide_data = [
            mx.io.DataDesc(name=data_names[0],
                           shape=(batch_size,) + source_image_size,  # "NCHW"
                           layout=C.BATCH_MAJOR_IMAGE)
        ]
        batch = mx.io.DataBatch([data_nd], None,
                                pad=0, index=None,
                                provide_data=provide_data)
        # Inference & tests
        module.forward(batch)
        feats = module.get_outputs()[0].asnumpy()
        assert feats.shape == (batch_size, encoded_seq_len, num_embed)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Code for inference/captioning.
"""
import logging
import os
from typing import List, Optional, Tuple

import mxnet as mx

from . import utils as utils_image
from .. import constants as C
from .. import lexical_constraints as constrained
from .. import model
from .. import utils
from .. import vocab
from ..inference import InferenceModel, Translator, \
    TranslatorInput, TranslatorOutput, models_max_input_output_length

logger = logging.getLogger(__name__)


class ImageInferenceModel(InferenceModel):
    """
    ImageInferenceModel is a InferenceModel that supports image models as encoders.
    """

    def __init__(self, input_size, **kwargs) -> None:
        super().__init__(**kwargs)
        self.input_size = input_size

    def _get_encoder_data_shapes(self, bucket_key: int = None) -> List[mx.io.DataDesc]:
        """
        Returns data shapes of the encoder module.

        :param bucket_key: Maximum input length.
        :return: List of data descriptions.
        """
        return [mx.io.DataDesc(name=C.SOURCE_NAME,
                               shape=(self.batch_size, ) + self.input_size,
                               layout=C.BATCH_MAJOR_IMAGE)]

    @property
    def max_supported_seq_len_source(self) -> Optional[int]:
        """ If not None this is the maximally supported source length during inference (hard constraint). """
        return None


class ImageCaptioner(Translator):
    """
    ImageCaptioner uses one or several models to output captions.
    It holds references to vocabularies to takes care of encoding input strings as word ids and conversion
    of target ids into a caption string.

    :param context: MXNet context to bind modules to.
    :param ensemble_mode: Ensemble mode: linear or log_linear combination.
    :param length_penalty: Length penalty instance.
    :param beam_prune: Beam pruning difference threshold.
    :param beam_search_stop: The stopping criterium.
    :param models: List of models.
    :param vocab_target: Target vocabulary.
    :param restrict_lexicon: Top-k lexicon to use for target vocabulary restriction.
    :param source_image_size: Shape of the image, input of the net
    :param source_root: Root where the images are stored
    :param use_feature_loader: Use precomputed features
    :param store_beam: If True, store the beam search history and return it in the TranslatorOutput.
    :param strip_unknown_words: If True, removes any <unk> symbols from outputs.
    """

    def __init__(self,
                 source_image_size: tuple,
                 source_root: str,
                 use_feature_loader: bool,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.source_image_size = source_image_size
        self.source_root = source_root
        self.use_feature_loader = use_feature_loader
        if self.use_feature_loader:
            self.data_loader = utils_image.load_features
        else:
            self.data_loader = utils_image.load_preprocess_images

    @staticmethod
    def make_input(sentence_id: int, sentence: str) -> TranslatorInput:
        """
        Returns TranslatorInput from input_string

        :param sentence_id: Input image id.
        :param sentence: Input image path.
        :return: Input for translate method.
        """
        return TranslatorInput(sentence_id=sentence_id, tokens=[sentence], factors=None)

    def translate(self, trans_inputs: List[TranslatorInput]) -> List[TranslatorOutput]:
        """
        Batch-translates a list of TranslatorInputs, returns a list of TranslatorOutputs.
        Splits oversized sentences to sentence chunks of size less than max_input_length.

        :param trans_inputs: List of TranslatorInputs as returned by make_input().
        :return: List of translation results.
        """

        # translate in batch-sized blocks over input chunks
        translations = []
        for batch_id, batch in enumerate(utils.grouper(trans_inputs, self.batch_size)):
            logger.debug("Translating batch %d", batch_id)
            # underfilled batch will be filled to a full batch size with copies of the 1st input
            rest = self.batch_size - len(batch)
            if rest > 0:
                logger.debug("Extending the last batch to the full batch size (%d)", self.batch_size)
                batch = batch + [batch[0]] * rest
            paths = []
            for b in batch:
                path = b.tokens[0]
                if self.source_root is not None:
                    path = os.path.join(self.source_root, path)
                paths.append(path)
            batch_translations = self._translate_nd(*self._get_inference_input(paths))
            # truncate to remove filler translations
            if rest > 0:
                batch_translations = batch_translations[:-rest]
            translations.extend(batch_translations)

        # Concatenate results
        results = []
        for trans_input, translation in zip(trans_inputs, translations):
            results.append(self._make_result(trans_input, translation))

        return results

    def _get_inference_input(self, image_paths: List[str]) -> Tuple[mx.nd.NDArray, int, List[Optional[constrained.RawConstraintList]]]:
        """
        Returns NDArray of images and corresponding bucket_key.

        :param image_paths: lists of image paths.
        :return NDArray of images paths, bucket key, a list of raw constraint lists.
        """
        ## TODO(bazzanil): support constraints
        raw_constraints = [None for x in range(self.batch_size)]  # type: List[Optional[constrained.RawConstraintList]]
        images = self.data_loader(image_paths, self.source_image_size)
        return mx.nd.array(images), 0, raw_constraints


def load_models(context: mx.context.Context,
                max_input_len: Optional[int],
                beam_size: int,
                batch_size: int,
                model_folders: List[str],
                checkpoints: Optional[List[int]] = None,
                softmax_temperature: Optional[float] = None,
                max_output_length_num_stds: int = C.DEFAULT_NUM_STD_MAX_OUTPUT_LENGTH,
                decoder_return_logit_inputs: bool = False,
                cache_output_layer_w_b: bool = False,
                source_image_size: tuple = None,
                forced_max_output_len: Optional[int] = None) -> Tuple[List[ImageInferenceModel],
                                                               List[vocab.Vocab],
                                                               vocab.Vocab]:
    """
    Loads a list of models for inference.

    :param context: MXNet context to bind modules to.
    :param max_input_len: Maximum input length.
    :param beam_size: Beam size.
    :param batch_size: Batch size.
    :param model_folders: List of model folders to load models from.
    :param checkpoints: List of checkpoints to use for each model in model_folders. Use None to load best checkpoint.
    :param softmax_temperature: Optional parameter to control steepness of softmax distribution.
    :param max_output_length_num_stds: Number of standard deviations to add to mean target-source length ratio
           to compute maximum output length.
    :param decoder_return_logit_inputs: Model decoders return inputs to logit computation instead of softmax over target
                                        vocabulary.  Used when logits/softmax are handled separately.
    :param cache_output_layer_w_b: Models cache weights and biases for logit computation as NumPy arrays (used with
                                   restrict lexicon).
    :param source_image_size: Size of the image to resize to. Used only for the image-text models
    :param forced_max_output_len: An optional overwrite of the maximum out length.
    :return: List of models, source vocabulary, target vocabulary, source factor vocabularies.
    """
    models = []  # type: List[InferenceModel]
    target_vocabs = []  # type: List[vocab.Vocab]

    if checkpoints is None:
        checkpoints = [None] * len(model_folders)

    for model_folder, checkpoint in zip(model_folders, checkpoints):
        target_vocabs.append(vocab.vocab_from_json(os.path.join(model_folder, C.VOCAB_TRG_NAME)))

        model_version = utils.load_version(os.path.join(model_folder, C.VERSION_NAME))
        logger.info("Model version: %s", model_version)
        utils.check_version(model_version)
        model_config = model.SockeyeModel.load_config(os.path.join(model_folder, C.CONFIG_NAME))

        if checkpoint is None:
            params_fname = os.path.join(model_folder, C.PARAMS_BEST_NAME)
        else:
            params_fname = os.path.join(model_folder, C.PARAMS_NAME % checkpoint)

        inference_model = ImageInferenceModel(config=model_config,
                                         params_fname=params_fname,
                                         context=context,
                                         beam_size=beam_size,
                                         batch_size=batch_size,
                                         softmax_temperature=softmax_temperature,
                                         decoder_return_logit_inputs=decoder_return_logit_inputs,
                                         cache_output_layer_w_b=cache_output_layer_w_b,
                                         input_size=source_image_size,
                                         forced_max_output_len=forced_max_output_len)

        models.append(inference_model)

    # set a common max_output length for all models.
    max_input_len, get_max_output_length = models_max_input_output_length(models,
                                                                          max_output_length_num_stds,
                                                                          max_input_len,
                                                                          forced_max_output_len=forced_max_output_len)

    for inference_model in models:
        inference_model.initialize(max_input_len, get_max_output_length)

    return models, None, target_vocabs[0]


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Implements a thin wrapper around ImageCaptioner to compute BLEU scores on
(a sample of) validation data during training.
"""
import logging
import os
import time
from typing import Dict, Optional

from .. import inference
from . import inference as inference_image
from .. import constants as C
from .. import data_io
from .. import evaluate
from .. import output_handler
from ..checkpoint_decoder import CheckpointDecoder

logger = logging.getLogger(__name__)


class CheckpointDecoderImageModel(CheckpointDecoder):
    """
    Decodes a (random sample of a) dataset using parameters at given checkpoint
    and computes BLEU against references.

    :param source_image_size: Size of the image feed into the net.
    :param image_root: Root where the images are stored.
    :param max_output_length: Max length of the generated sentence.
    :param use_feature_loader: If True, features are loaded instead of images.
    :param kwargs: Arguments passed to `sockeye.checkpoint_decoder.CheckpointDecoder`.
    """

    def __init__(self,
                 source_image_size: tuple,
                 image_root: str,
                 max_output_length: int = 50,
                 use_feature_loader: bool = False,
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.source_image_size = source_image_size
        self.image_root = image_root
        self.max_output_length = max_output_length
        self.use_feature_loader = use_feature_loader

    def decode_and_evaluate(self,
                            checkpoint: Optional[int] = None,
                            output_name: str = os.devnull) -> Dict[str, float]:
        """
        Decodes data set and evaluates given a checkpoint.

        :param checkpoint: Checkpoint to load parameters from.
        :param output_name: Filename to write translations to. Defaults to /dev/null.
        :return: Mapping of metric names to scores.
        """

        models, vocab_source, vocab_target = inference_image.load_models(
            context=self.context,
            max_input_len=self.max_input_len,
            beam_size=self.beam_size,
            batch_size=self.batch_size,
            model_folders=[self.model],
            checkpoints=[checkpoint],
            softmax_temperature=self.softmax_temperature,
            max_output_length_num_stds=self.max_output_length_num_stds,
            source_image_size=tuple(self.source_image_size),
            forced_max_output_len=self.max_output_length
        )
        translator = inference_image.ImageCaptioner(context=self.context,
                                              ensemble_mode=self.ensemble_mode,
                                              bucket_source_width=0,
                                              length_penalty=inference.LengthPenalty(
                                                  self.length_penalty_alpha,
                                                  self.length_penalty_beta),
                                              beam_prune=0.0,
                                              beam_search_stop='all',
                                              models=models,
                                              source_vocabs=None,
                                              target_vocab=vocab_target,
                                              restrict_lexicon=None,
                                              store_beam=False,
                                              source_image_size=tuple(
                                                  self.source_image_size),
                                              source_root=self.image_root,
                                              use_feature_loader=self.use_feature_loader)

        trans_wall_time = 0.0
        translations = []
        with data_io.smart_open(output_name, 'w') as output:
            handler = output_handler.StringOutputHandler(output)
            tic = time.time()
            trans_inputs = []  # type: List[inference.TranslatorInput]
            for i, inputs in enumerate(self.inputs_sentences):
                trans_inputs.append(
                    inference.make_input_from_multiple_strings(i, inputs))
            trans_outputs = translator.translate(trans_inputs)
            trans_wall_time = time.time() - tic
            for trans_input, trans_output in zip(trans_inputs, trans_outputs):
                handler.handle(trans_input, trans_output)
                translations.append(trans_output.translation)
        avg_time = trans_wall_time / len(self.target_sentences)

        # TODO(fhieber): eventually add more metrics (METEOR etc.)
        return {C.BLEU_VAL: evaluate.raw_corpus_bleu(hypotheses=translations,
                                                     references=self.target_sentences,
                                                     offset=0.01),
                C.CHRF_VAL: evaluate.raw_corpus_chrf(hypotheses=translations,
                                                     references=self.target_sentences),
                C.AVG_TIME: avg_time,
                C.DECODING_TIME: trans_wall_time}



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Encoders for sequence-to-sequence models.
"""
import logging
import mxnet as mx
from typing import List, Tuple

from .. import constants as C
from ..config import Config
from ..encoder import EncoderSequence, Encoder
from ..encoder import get_positional_embedding

logger = logging.getLogger(__name__)


class ImageLoadedCnnEncoderConfig(Config):
    """
    Image cnn encoder configuration. The symbolic model is loaded from disk.

    :param model_path: Path where the json file is stored.
    :param epoch: Epoch of the pre-trained model.
    :param layer_name: Name of the layer of the loaded symbol to get the encoding from.
    :param encoded_seq_len: Size of the feature layer. If the layer is a conv layer.
        encoded_seq_len should be equal to the height*width of the convolutional map,
        the number of kernel is not considered.
    :param num_embed: Number of hiddens to project the local features to.
    :param no_global_descriptor: By default the global visual feature (spatial avg of conv map)
        is concatenated to the local visual features (conv map). This option disables the use of
        the global descriptor, such that only the local ones are used.
    :param number_of_kernels: If using preextracted features, we need to know the number of dim of the features.
    :param positional_embedding_type: Which king of positional embeddingm if any.
    :param preextracted_features: Turn to bool if you preextracted featured from existing model.
    """

    def __init__(self,
                 model_path: str,
                 epoch: int,
                 layer_name: str,
                 encoded_seq_len: int,
                 num_embed: int,
                 no_global_descriptor: bool = True,
                 number_of_kernels: int = None,
                 positional_embedding_type: str = "",
                 preextracted_features: bool = False) -> None:
        super().__init__()
        self.model_path = model_path
        self.layer_name = layer_name
        self.epoch = epoch
        self.encoded_seq_len = encoded_seq_len
        self.num_embed = num_embed
        self.no_global_descriptor = no_global_descriptor
        self.number_of_kernels = number_of_kernels
        self.positional_embedding_type = positional_embedding_type
        self.preextracted_features = preextracted_features


def get_image_cnn_encoder(config: ImageLoadedCnnEncoderConfig) -> 'Encoder':
    """
    Creates a image encoder.

    :param config: Configuration for image encoder.
    :return: Encoder instance.
    """

    encoders = list()  # type: List[Encoder]
    max_seq_len = config.encoded_seq_len
    if not config.no_global_descriptor:
        max_seq_len += 1
    encoders.append(get_positional_embedding(config.positional_embedding_type,
                                             config.num_embed,
                                             max_seq_len=max_seq_len,
                                             fixed_pos_embed_scale_up_input=False,
                                             fixed_pos_embed_scale_down_positions=True,
                                             prefix=C.SOURCE_POSITIONAL_EMBEDDING_PREFIX))
    encoders.append(ImageLoadedCnnEncoder(config=config))
    return EncoderSequence(encoders)


class ImageLoadedCnnEncoder(Encoder):
    """
    Image cnn encoder. The model is loaded from disk.

    :param config: Image cnn encoder config.
    :param prefix: Name prefix for symbols of this encoder.
    """

    def __init__(self,
                 config: ImageLoadedCnnEncoderConfig,
                 prefix: str = C.CHAR_SEQ_ENCODER_PREFIX) -> None:
        self.model_path = config.model_path
        self.layer_name = config.layer_name
        self.epoch = config.epoch
        self.encoded_seq_len = config.encoded_seq_len
        self.num_embed = config.num_embed
        self.no_global_descriptor = config.no_global_descriptor
        self.preextracted_features = config.preextracted_features
        if not self.preextracted_features:
            sym, args, auxs = mx.model.load_checkpoint(self.model_path, self.epoch)
            # get layers up to layer_name
            all_layers = sym.get_internals()
            try:
                self.sym = all_layers[self.layer_name + "_output"]
            except ValueError:
                raise ValueError("Layer {} not found in the architecure located at "
                               "{}. Make sure that you choose an existing layer."\
                               .format(self.layer_name, self.model_path))
            # throws away fc weights
            self.args = dict({k: args[k] for k in args if 'fc1' not in k})
            self.auxs = auxs
            self.n_kernels = self.args[self.layer_name + "_weight"].shape[0]
            # "rename" input
            self.input = mx.sym.Variable(name=C.SOURCE_NAME)
            self.sym = self.sym(data=self.input)
        else:
            self.args = {}
            self.auxs = {}
            self.n_kernels = config.number_of_kernels
            self.sym = mx.sym.Variable(name=C.SOURCE_NAME)
        self.names = ["local_image_encoding_weight"]
        self.other_weights = {self.names[0]: mx.sym.Variable(self.names[0])}
        if not self.no_global_descriptor:
            self.names.append("global_image_encoding_weight")
            self.other_weights[self.names[1]] = mx.sym.Variable(self.names[1])
            self.encoded_seq_len += 1
        # output
        self.output_dim = self.num_embed

    def encode(self,
               data: mx.sym.Symbol,
               data_length: mx.sym.Symbol,
               seq_len: int) -> Tuple[mx.sym.Symbol, mx.sym.Symbol, int]:
        """
        Encodes data given sequence lengths of individual examples and maximum sequence length.

        :param data: Ignored. Assume that the input is the image.
        :param data_length: Vector with sequence lengths.
        :param seq_len: Maximum sequence length.
        :return: Encoded versions of input data data, data_length, seq_len.
        """

        # (batch, n_kernels, height, width) -> (batch, width, height, n_kernels)
        embedding = mx.sym.swapaxes(data=self.sym, dim1=1, dim2=3)
        # (batch, width, height, n_kernels) -> (batch, height, width, n_kernels)
        embedding = mx.sym.swapaxes(data=embedding, dim1=1, dim2=2)
        # (batch, height, width, n_kernels) -> (batch, height*width, n_kernels)
        embedding = mx.sym.Reshape(data=embedding, shape=(0, -3, self.n_kernels))
        # Feature projection layer: (batch, height*width, num_embed)
        embedding = mx.sym.FullyConnected(data=embedding, weight=self.other_weights[self.names[0]],
                                   num_hidden=self.num_embed, no_bias=True, flatten=False)
        embedding = mx.sym.Activation(data=embedding, act_type='relu')

        # Visual global description: average pooling
        if not self.no_global_descriptor:
            glob_embedding = mx.sym.mean(data=embedding, axis=1) # (batch, n_kernels)
            glob_embedding = mx.sym.FullyConnected(data=glob_embedding, weight=self.other_weights[self.names[1]],
                                       num_hidden=self.num_embed, no_bias=True)
            glob_embedding = mx.sym.Activation(data=glob_embedding, act_type='relu')
            glob_embedding = mx.sym.expand_dims(glob_embedding, axis=1)
            # Concatenate embeddings with global embedding: (batch, height*width+1, num_embed)
            embedding = mx.sym.concat(embedding, glob_embedding, dim=1, name="local_global_image_embedding")

        # Symbol to infer axis 1 dimension
        d = mx.sym.slice_axis(data=embedding, axis=2, begin=0, end=1)  # (batch, height*width, num_embed)
        d = mx.sym.clip(data=d, a_min=1.0, a_max=1.0)  # matrix of all ones
        encoded_data_length = mx.sym.sum(mx.sym.broadcast_equal(d, mx.sym.ones((1,))), axis=1)  # (batch, 1)
        encoded_data_length = mx.sym.reshape(data=encoded_data_length, shape= (-1,)) # (batch, )

        return embedding, encoded_data_length, self.encoded_seq_len

    def get_params(self):
        """
        Get the parameters of the pre-trained networks.

        :return: Tuple of arguments and auxiliaries
        """
        return self.args, self.auxs

    def get_num_hidden(self) -> int:
        """
        Return the representation size of this encoder.
        """
        return self.output_dim

    def get_encoded_seq_len(self, seq_len: int) -> int:
        """
        :return: The size of the encoded sequence.
        """
        return self.encoded_seq_len

    def get_initializers(self) -> List[Tuple[str, mx.init.Initializer]]:
        """
        Get the initializers of the network, considering the pretrained models.

        :return: List of tuples (string name, mxnet initializer)
        """
        patterns_vals = []
        # Load from args/auxs
        for k in self.args.keys():
            patterns_vals.append((k, mx.init.Load({k: self.args[k]})))
        for k in self.auxs.keys():
            patterns_vals.append((k, mx.init.Load({k: self.auxs[k]})))
        # Initialize
        for k in self.names:
            patterns_vals.append((k, mx.init.Xavier(rnd_type='uniform', factor_type='avg', magnitude=3)))

        return patterns_vals

    def get_fixed_param_names(self) -> List[str]:
        """
        Get the fixed params of the network.

        :return: List of strings, names of the layers
        """
        args = set(self.args.keys()) | set(self.auxs.keys())

        return list(args & set(self.sym.list_arguments()))


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
from unittest.mock import Mock

import mxnet as mx
import numpy as np
import pytest
from math import ceil

import sockeye.constants as C
import sockeye.data_io
import sockeye.inference
from sockeye.utils import SockeyeError
from sockeye.lexical_constraints import init_batch, get_bank_sizes, topk, ConstrainedHypothesis

BOS_ID = 2
EOS_ID = 3

def mock_translator(num_source_factors: int):
    t_mock = Mock(sockeye.inference.Translator)
    t_mock.num_source_factors = num_source_factors
    return t_mock


"""
Test how the banks are allocated. Given a number of constraints (C), the beam size (k)
a number of candidates for each bank [0..C], return the allocation of the k spots of the
beam to the banks.
"""
@pytest.mark.parametrize("num_constraints, beam_size, counts, expected_allocation",
                         [
                             # no constraints: allocate all items to bin 0
                             (0, 5, [5], [5]),
                             # 1 constraints, but no candidates, so 0 alloc
                             (1, 5, [5,0], [5,0]),
                             # 1 constraint, but 1 candidate, so 1 alloc
                             (1, 5, [5,1], [4,1]),
                             # 1 constraint, > k candidates for each (impossible, but ok), even alloc, extra goes to last
                             (1, 5, [10,10], [2,3]),
                             # 1 constraint, > k candidates for each (impossible, but ok), even alloc, extra goes to last
                             (1, 5, [1,10], [1,4]),
                             # 2 constraints, no candidates
                             (2, 5, [5,0,0], [5,0,0]),
                             # 2 constraints, some candidates
                             (2, 10, [5,0,7], [5,0,5]),
                             # 2 constraints, some candidates
                             (2, 10, [5,1,7], [5,1,4]),
                             # more constraints than beam spots: allocate to last two spots
                             (3, 2, [1,0,1,1], [0,0,1,1]),
                             # more constraints than beam spots: slots allocated empirically
                             (3, 2, [1,0,1,0], [1,0,1,0]),
                             # more constraints than beam spots: all slots to last bank
                             (3, 2, [4,2,2,3], [0,0,0,2]),
                         ])
def test_constraints_bank_allocation(num_constraints, beam_size, counts, expected_allocation):
    allocation = get_bank_sizes(num_constraints, beam_size, counts)
    assert sum(allocation) == beam_size
    assert allocation == expected_allocation


"""
Make sure the internal representation is correct.
For the internal representation, the list of phrasal constraints is concatenated, and then
a parallel array is used to mark which words are part of a phrasal constraint.
"""
@pytest.mark.parametrize("raw_constraints, internal_constraints, internal_is_sequence",
                         [
                             # No constraints
                             ([], [], []),
                             # One single-word constraint
                             ([[17]], [17], [0]),
                             # Multiple multiple-word constraints.
                             ([[11, 12], [13, 14]], [11, 12, 13, 14], [True, False, True, False]),
                             # Multiple constraints
                             ([[11, 12, 13], [14], [15]], [11, 12, 13, 14, 15], [True, True, False, False, False]),
                         ])
def test_constraints_repr(raw_constraints, internal_constraints, internal_is_sequence):
    hyp = ConstrainedHypothesis(raw_constraints, EOS_ID)
    assert hyp.constraints == internal_constraints
    assert hyp.is_sequence == internal_is_sequence


"""
Tests many of the ConstrainedHypothesis functions.
"""
@pytest.mark.parametrize("raw_constraints, met, unmet",
                         [
                             # No constraints
                             ([], [], []),
                             # Single simple unmet constraint
                             ([[17]], [], [17]),
                             # Single simple met constraint
                             ([[17]], [17], []),
                             # Met first word of a phrasal constraint, return just next word of phrasal
                             ([[11, 12], [13, 14]], [11], [12, 13, 14]),
                             # Completed phrase, have only single-word ones
                             ([[11, 12, 13], [14], [15]], [11, 12, 13], [14, 15]),
                             # Same word twice
                             ([[11], [11]], [], [11, 11]),
                             ])
def test_constraints_logic(raw_constraints, met, unmet):
    hyp = ConstrainedHypothesis(raw_constraints, EOS_ID)
    # record these ones as met
    for word_id in met:
        hyp = hyp.advance(word_id)

    assert hyp.num_needed() == len(unmet)
    assert hyp.finished() == (len(unmet) == 0)
    assert hyp.is_valid(EOS_ID) == (hyp.finished() or (len(unmet) == 1 and EOS_ID in unmet))


"""
Test the allowed() function, which returns the set of unmet constraints that can be generated.
When inside a phrase, this is only the next word of the phrase. Otherwise, it is all unmet constraints.
"""
@pytest.mark.parametrize("raw_constraints, met, allowed",
                         [
                             # No constraints
                             ([], [], []),
                             # Single simple unmet constraint
                             ([[17]], [], [17]),
                             # Single simple met constraint
                             ([[17]], [17], []),
                             # Met first word of a phrasal constraint, return just next word of phrasal
                             ([[11, 12], [13, 14]], [11], [12]),
                             # Completed phrase, have only single-word ones
                             ([[11, 12, 13], [14], [15]], [11, 12, 13], [14, 15]),
                             # Same word twice, nothing met, return
                             ([[11], [11]], [], [11]),
                             # Same word twice, met, still returns once
                             ([[11], [11]], [11], [11]),
                             # Same word twice, met twice
                             ([[11], [11]], [11, 11], []),
                             # EOS, allowed
                             ([[42, EOS_ID]], [42], [EOS_ID]),
                             # EOS, not allowed
                             ([[42, EOS_ID]], [], [42]),
                         ])
def test_constraints_allowed(raw_constraints, met, allowed):
    hyp = ConstrainedHypothesis(raw_constraints, EOS_ID)
    # record these ones as met
    for word_id in met:
        hyp = hyp.advance(word_id)

    assert hyp.allowed() == set(allowed)
    assert hyp.num_met() == len(met)
    assert hyp.num_needed() == hyp.size() - hyp.num_met()





"""
Ensures that batches are initialized correctly.
Each line here is a tuple containing a list (for each sentence in the batch) of RawConstraintLists,
which are lists of list of integer IDs representing the constraints for the sentence.
"""
@pytest.mark.parametrize("raw_constraint_lists",
                         [ ([None, None, None, None]),
                           ([[[17]], None]),
                           ([None, [[17]]]),
                           ([[[17], [11, 12]], [[17]], None]),
                           ([None, [[17], [11, 12]], [[17]], None]),
                         ])
def test_constraints_init_batch(raw_constraint_lists):
    beam_size = 4  # arbitrary

    constraints = init_batch(raw_constraint_lists, beam_size, BOS_ID, EOS_ID)
    assert len(raw_constraint_lists) * beam_size == len(constraints)

    # Iterate over sentences in the batch
    for raw_constraint_list, constraint in zip(raw_constraint_lists, constraints[::beam_size]):
        if raw_constraint_list is None:
            assert constraint is None
        else:
            # The number of constraints is the sum of the length of the lists in the raw constraint list
            assert constraint.size() == sum([len(phr) for phr in raw_constraint_list])

            # No constraints are met unless the start_id happened to be at the start of a constraint
            num_met = 1 if any([phr[0] == BOS_ID for phr in raw_constraint_list]) else 0
            assert constraint.num_met() == num_met



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Image captioning CLI.
"""
import argparse
import os
import tempfile
from contextlib import ExitStack

import mxnet as mx

from . import arguments as arguments_image
from . import inference as inference_image
from .train import read_feature_shape
from .. import arguments
from .. import constants as C
from .. import inference
from .. import output_handler
from ..image_captioning import utils
from ..image_captioning.extract_features import get_pretrained_net, \
    batching, read_list_file, extract_features_forward
from ..lexicon import TopKLexicon
from ..log import setup_main_logger
from ..translate import read_and_translate, _setup_context
from ..utils import check_condition
from ..utils import log_basic_info

logger = setup_main_logger(__name__, file_logging=False)


def get_pretrained_caption_net(args: argparse.Namespace,
                               context: mx.Context,
                               image_preextracted_features: bool) -> inference_image.ImageCaptioner:
    models, source_vocabs, target_vocab = inference_image.load_models(
        context=context,
        max_input_len=args.max_input_len,
        beam_size=args.beam_size,
        batch_size=args.batch_size,
        model_folders=args.models,
        checkpoints=args.checkpoints,
        softmax_temperature=args.softmax_temperature,
        max_output_length_num_stds=args.max_output_length_num_stds,
        decoder_return_logit_inputs=args.restrict_lexicon is not None,
        cache_output_layer_w_b=args.restrict_lexicon is not None,
        source_image_size=tuple(args.feature_size),
        forced_max_output_len=args.max_output_length
    )
    restrict_lexicon = None  # type: TopKLexicon
    store_beam = args.output_type == C.OUTPUT_HANDLER_BEAM_STORE
    if args.restrict_lexicon:
        restrict_lexicon = TopKLexicon(source_vocabs, target_vocab)
        restrict_lexicon.load(args.restrict_lexicon)

    translator = inference_image.ImageCaptioner(context=context,
                                                ensemble_mode=args.ensemble_mode,
                                                bucket_source_width=0,
                                                length_penalty=inference.LengthPenalty(
                                                    args.length_penalty_alpha,
                                                    args.length_penalty_beta),
                                                beam_prune=args.beam_prune,
                                                beam_search_stop=args.beam_search_stop,
                                                models=models,
                                                source_vocabs=None,
                                                target_vocab=target_vocab,
                                                restrict_lexicon=restrict_lexicon,
                                                store_beam=store_beam,
                                                strip_unknown_words=args.strip_unknown_words,
                                                source_image_size=tuple(
                                                    args.feature_size),
                                                source_root=args.source_root,
                                                use_feature_loader=image_preextracted_features)
    return translator


def _extract_features(args, context):
    image_list = read_list_file(args.input)
    image_model, _ = get_pretrained_net(args, context)
    output_root = tempfile.mkdtemp()
    output_file = os.path.join(output_root, "input.features")
    with open(output_file, "w") as fout:
        for i, im in enumerate(batching(image_list, args.batch_size)):
            feats, out_names = extract_features_forward(im, image_model,
                                                        args.source_root,
                                                        output_root,
                                                        args.batch_size,
                                                        args.source_image_size,
                                                        context)
            # Save to disk
            out_file_names = utils.save_features(out_names, feats)
            # Write to output file
            out_file_names = map(lambda x: os.path.basename(x) + "\n",
                                 out_file_names)
            fout.writelines(out_file_names)
    return output_root, output_file, tuple(feats.shape[1:])


def main():
    params = arguments.ConfigArgumentParser(description='Image Captioning CLI')
    arguments_image.add_image_caption_cli_args(params)
    args = params.parse_args()
    image_preextracted_features = not args.extract_image_features

    if args.output is not None:
        global logger
        logger = setup_main_logger(__name__,
                                   console=not args.quiet,
                                   file_logging=True,
                                   path="%s.%s" % (args.output, C.LOG_NAME))

    if args.checkpoints is not None:
        check_condition(len(args.checkpoints) == len(args.models),
                        "must provide checkpoints for each model")

    log_basic_info(args)

    out_handler = output_handler.get_output_handler(args.output_type,
                                                    args.output,
                                                    args.sure_align_threshold)

    with ExitStack() as exit_stack:
        context = _setup_context(args, exit_stack)

        if not image_preextracted_features:
            # Extract features and override input and source_root with tmp location of features
            args.source_root, args.input, args.feature_size = _extract_features(
                args, context)
            image_preextracted_features = True  # now we extracted features
        else:  # Read feature size from disk
            _, args.feature_size = read_feature_shape(args.source_root)

        translator = get_pretrained_caption_net(args, context,
                                                image_preextracted_features)

        read_and_translate(translator=translator,
                           output_handler=out_handler,
                           chunk_size=args.chunk_size,
                           input_file=args.input)


if __name__ == '__main__':
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Training CLI for image captioning.
"""
import argparse
import json
import os
import pickle
from contextlib import ExitStack
from typing import cast, Dict, List, Tuple, Optional

import mxnet as mx
import numpy as np

from ..config import Config
from ..log import setup_main_logger
from ..train import check_resume, check_arg_compatibility, \
    determine_context, create_decoder_config, \
    create_optimizer_config, create_training_model
from ..utils import check_condition
# Sockeye captioner
from . import arguments as arguments_image
from . import checkpoint_decoder
from . import data_io as data_io_image
from . import encoder as encoder_image
from .. import constants as C
# Sockeye
from .. import arguments
from .. import data_io
from .. import encoder
from .. import loss
from .. import model
from .. import training
from .. import utils
from .. import vocab

# Temporary logger, the real one (logging to a file probably, will be created in the main function)
logger = setup_main_logger(__name__, file_logging=False, console=True)


def read_feature_shape(path):
    shape_file = os.path.join(path, "image_feature_sizes.pkl")
    with open(shape_file, "rb") as fout:
        shapes = pickle.load(fout)
    return shapes["image_shape"], shapes["features_shape"]


def create_checkpoint_decoder(args: argparse.Namespace,
                              exit_stack: ExitStack,
                              train_context: List[mx.Context]) -> Optional[checkpoint_decoder.CheckpointDecoder]:
    """
    Returns a checkpoint decoder or None.

    :param args: Arguments as returned by argparse.
    :param exit_stack: An ExitStack from contextlib.
    :param train_context: Context for training.
    :return: A CheckpointDecoder if --decode-and-evaluate != 0, else None.
    """
    sample_size = args.decode_and_evaluate
    if args.optimized_metric == C.BLEU and sample_size == 0:
        logger.info("You chose BLEU as the optimized metric, will turn on BLEU monitoring during training. "
                    "To control how many validation sentences are used for calculating bleu use "
                    "the --decode-and-evaluate argument.")
        sample_size = -1

    if sample_size == 0:
        return None

    if args.use_cpu or args.decode_and_evaluate_use_cpu:
        context = mx.cpu()
    elif args.decode_and_evaluate_device_id is not None:
        # decode device is defined from the commandline
        num_gpus = utils.get_num_gpus()
        check_condition(num_gpus >= 1,
                        "No GPUs found, consider running on the CPU with --use-cpu "
                        "(note: check depends on nvidia-smi and this could also mean that the nvidia-smi "
                        "binary isn't on the path).")

        if args.disable_device_locking:
            context = utils.expand_requested_device_ids([args.decode_and_evaluate_device_id])
        else:
            context = exit_stack.enter_context(utils.acquire_gpus([args.decode_and_evaluate_device_id],
                                                                  lock_dir=args.lock_dir))
        context = mx.gpu(context[0])

    else:
        # default decode context is the last training device
        context = train_context[-1]

    return checkpoint_decoder.CheckpointDecoderImageModel(context=context,
                                                inputs=[args.validation_source] + args.validation_source_factors,
                                                references=args.validation_target,
                                                model=args.output,
                                                sample_size=sample_size,
                                                source_image_size=args.source_image_size,
                                                image_root=args.validation_source_root,
                                                max_output_length=args.max_output_length,
                                                use_feature_loader=args.image_preextracted_features)


def create_data_iters_and_vocab(args: argparse.Namespace,
                                max_seq_len_source: int,
                                max_seq_len_target: int,
                                resume_training: bool,
                                output_folder: str) -> Tuple['data_io.BaseParallelSampleIter',
                                                             'data_io.BaseParallelSampleIter',
                                                             'data_io.DataConfig', Dict]:
    """
    Create the data iterators and the vocabularies.

    :param args: Arguments as returned by argparse.
    :param max_seq_len_source: Source maximum sequence length.
    :param max_seq_len_target: Target maximum sequence length.
    :param resume_training: Whether to resume training.
    :param output_folder: Output folder.
    :return: The data iterators (train, validation, config_data) as well as the source and target vocabularies.
    """

    _, num_words_target = args.num_words
    _, word_min_count_target = args.word_min_count
    batch_num_devices = 1 if args.use_cpu else sum(-di if di < 0 else 1 for di in args.device_ids)
    batch_by_words = args.batch_type == C.BATCH_TYPE_WORD

    either_raw_or_prepared_error_msg = "Either specify a raw training corpus with %s or a preprocessed corpus " \
                                       "with %s." % (C.TRAINING_ARG_TARGET,
                                                     C.TRAINING_ARG_PREPARED_DATA)
    # Note: ignore args.prepared_data for the moment
    utils.check_condition(args.prepared_data is None and args.target is not None,
                          either_raw_or_prepared_error_msg)

    if resume_training:
        # Load the existing vocab created when starting the training run.
        target_vocab = vocab.vocab_from_json(os.path.join(output_folder, C.VOCAB_TRG_NAME))

        # Recover the vocabulary path from the existing config file:
        data_info = cast(data_io.DataInfo, Config.load(os.path.join(output_folder, C.DATA_INFO)))
        target_vocab_path = data_info.target_vocab
    else:
        # Load vocab:
        target_vocab_path = args.target_vocab
        # Note: We do not care about the source vocab for images, that is why some inputs are mocked
        target_vocab = vocab.load_or_create_vocab(data=args.target,
                                                  vocab_path=target_vocab_path,
                                                  num_words=num_words_target,
                                                  word_min_count=word_min_count_target)

    train_iter, validation_iter, config_data, data_info = data_io_image.get_training_image_text_data_iters(
        source_root=args.source_root,
        source=os.path.abspath(args.source),
        target=os.path.abspath(args.target),
        validation_source_root=args.validation_source_root,
        validation_source=os.path.abspath(args.validation_source),
        validation_target=os.path.abspath(args.validation_target),
        vocab_target=target_vocab,
        vocab_target_path=target_vocab_path,
        batch_size=args.batch_size,
        batch_by_words=batch_by_words,
        batch_num_devices=batch_num_devices,
        source_image_size=args.source_image_size,
        fill_up=args.fill_up,
        max_seq_len_target=max_seq_len_target,
        bucketing=not args.no_bucketing,
        bucket_width=args.bucket_width,
        use_feature_loader=args.image_preextracted_features,
        preload_features=args.load_all_features_to_memory
    )

    data_info_fname = os.path.join(output_folder, C.DATA_INFO)
    logger.info("Writing data config to '%s'", data_info_fname)
    # Removing objects that cannot be saved:
    data_info.sources = None
    data_info.save(data_info_fname)

    return train_iter, validation_iter, config_data, target_vocab


def create_encoder_config(args: argparse.Namespace) -> Tuple[Config, int]:

    if args.encoder == C.IMAGE_PRETRAIN_TYPE:
        number_of_kernels = args.source_image_size[0]
        encoded_seq_len = np.prod(args.source_image_size[1:])
        config_encoder = encoder_image.ImageLoadedCnnEncoderConfig(model_path=args.image_encoder_model_path,
                                                          epoch=args.image_encoder_model_epoch,
                                                          layer_name=args.image_encoder_layer,
                                                          encoded_seq_len=encoded_seq_len,
                                                          num_embed=args.image_encoder_num_hidden,
                                                          no_global_descriptor=args.no_image_encoder_global_descriptor,
                                                          preextracted_features=args.image_preextracted_features,
                                                          number_of_kernels=number_of_kernels,
                                                          positional_embedding_type=args.image_positional_embedding_type)
        encoder_num_hidden = args.image_encoder_num_hidden
    else:
        raise ValueError("Image encoder must be provided. (current: {}, "
                         "expected: {})".format(args.encoder, C.ENCODERS))
    return config_encoder, encoder_num_hidden


def create_model_config(args: argparse.Namespace,
                        vocab_target_size: int,
                        max_seq_len_source: int,
                        max_seq_len_target: int,
                        config_data: data_io.DataConfig) -> model.ModelConfig:
    """
    Create a ModelConfig from the argument given in the command line.

    :param args: Arguments as returned by argparse.
    :param vocab_target_size: The size of the target vocabulary.
    :param max_seq_len_source: Maximum source sequence length.
    :param max_seq_len_target: Maximum target sequence length.
    :param config_data: Data config.
    :return: The model configuration.
    """
    num_embed_source, num_embed_target = args.num_embed
    _, embed_dropout_target = args.embed_dropout

    config_encoder, encoder_num_hidden = create_encoder_config(args)
    config_decoder = create_decoder_config(args, encoder_num_hidden, max_seq_len_source, max_seq_len_target)

    config_embed_source = encoder.PassThroughEmbeddingConfig()
    config_embed_target = encoder.EmbeddingConfig(vocab_size=vocab_target_size,
                                                  num_embed=num_embed_target,
                                                  dropout=embed_dropout_target)

    config_loss = loss.LossConfig(name=args.loss,
                                  vocab_size=vocab_target_size,
                                  normalization_type=args.loss_normalization_type,
                                  label_smoothing=args.label_smoothing)

    model_config = model.ModelConfig(config_data=config_data,
                                     vocab_source_size=0,
                                     vocab_target_size=vocab_target_size,
                                     config_embed_source=config_embed_source,
                                     config_embed_target=config_embed_target,
                                     config_encoder=config_encoder,
                                     config_decoder=config_decoder,
                                     config_loss=config_loss,
                                     weight_tying=args.weight_tying,
                                     weight_tying_type=args.weight_tying_type if args.weight_tying else None,
                                     weight_normalization=args.weight_normalization,
                                     lhuc=args.lhuc is not None)
    return model_config


def get_preinit_encoders(encoders: List[encoder.Encoder]) -> List[Tuple[str, mx.init.Initializer]]:
    """
    Get initializers from encoders. Some encoders might be initialized from pretrained models.

    :param encoders: List of encoders
    :return: The list of initializers
    """
    init = []
    for enc in encoders:
        if hasattr(enc, "get_initializers"):
            init.extend(enc.get_initializers())
    return init


def main():
    params = arguments.ConfigArgumentParser(description='Train Sockeye images-to-text models.')
    arguments_image.add_image_train_cli_args(params)
    args = params.parse_args()
    # TODO: make training compatible with full net
    args.image_preextracted_features = True  # override this for now

    utils.seedRNGs(args.seed)

    check_arg_compatibility(args)
    output_folder = os.path.abspath(args.output)
    resume_training = check_resume(args, output_folder)

    global logger
    logger = setup_main_logger(__name__,
                               file_logging=True,
                               console=not args.quiet, path=os.path.join(output_folder, C.LOG_NAME))
    utils.log_basic_info(args)
    with open(os.path.join(output_folder, C.ARGS_STATE_NAME), "w") as fp:
        json.dump(vars(args), fp)

    max_seq_len_source, max_seq_len_target = args.max_seq_len
    # The maximum length is the length before we add the BOS/EOS symbols
    max_seq_len_source = max_seq_len_source + C.SPACE_FOR_XOS
    max_seq_len_target = max_seq_len_target + C.SPACE_FOR_XOS
    logger.info("Adjusting maximum length to reserve space for a BOS/EOS marker. New maximum length: (%d, %d)",
                max_seq_len_source, max_seq_len_target)

    with ExitStack() as exit_stack:
        context = determine_context(args, exit_stack)

        # Read feature size
        if args.image_preextracted_features:
            _, args.source_image_size = read_feature_shape(args.source_root)

        train_iter, eval_iter, config_data, target_vocab = create_data_iters_and_vocab(
            args=args,
            max_seq_len_source=max_seq_len_source,
            max_seq_len_target=max_seq_len_target,
            resume_training=resume_training,
            output_folder=output_folder)
        max_seq_len_source = config_data.max_seq_len_source
        max_seq_len_target = config_data.max_seq_len_target

        # Dump the vocabularies if we're just starting up
        if not resume_training:
            vocab.vocab_to_json(target_vocab, os.path.join(output_folder, C.VOCAB_TRG_NAME))

        target_vocab_size = len(target_vocab)
        logger.info("Vocabulary sizes: target=%d", target_vocab_size)

        model_config = create_model_config(args=args,
                                           vocab_target_size=target_vocab_size,
                                           max_seq_len_source=max_seq_len_source, max_seq_len_target=max_seq_len_target,
                                           config_data=config_data)
        model_config.freeze()

        training_model = create_training_model(config=model_config,
                                               context=context,
                                               output_dir=output_folder,
                                               train_iter=train_iter,
                                               args=args)

        # Handle options that override training settings
        min_updates = args.min_updates
        max_updates = args.max_updates
        min_samples = args.min_samples
        max_samples = args.max_samples
        max_num_checkpoint_not_improved = args.max_num_checkpoint_not_improved
        min_epochs = args.min_num_epochs
        max_epochs = args.max_num_epochs
        if min_epochs is not None and max_epochs is not None:
            check_condition(min_epochs <= max_epochs,
                            "Minimum number of epochs must be smaller than maximum number of epochs")
        # Fixed training schedule always runs for a set number of updates
        if args.learning_rate_schedule:
            min_updates = None
            max_updates = sum(num_updates for (_, num_updates) in args.learning_rate_schedule)
            max_num_checkpoint_not_improved = -1
            min_samples = None
            max_samples = None
            min_epochs = None
            max_epochs = None

        # Get initialization from encoders (useful for pretrained models)
        extra_initializers = get_preinit_encoders(training_model.encoder.encoders)
        if len(extra_initializers)==0:
            extra_initializers = None

        trainer = training.EarlyStoppingTrainer(model=training_model,
                                                optimizer_config=create_optimizer_config(args, [1.0], extra_initializers),
                                                max_params_files_to_keep=args.keep_last_params,
                                                source_vocabs=[None],
                                                target_vocab=target_vocab)

        trainer.fit(train_iter=train_iter,
                    validation_iter=eval_iter,
                    early_stopping_metric=args.optimized_metric,
                    metrics=args.metrics,
                    checkpoint_frequency=args.checkpoint_frequency,
                    max_num_not_improved=max_num_checkpoint_not_improved,
                    min_samples=min_samples,
                    max_samples=max_samples,
                    min_updates=min_updates,
                    max_updates=max_updates,
                    min_epochs=min_epochs,
                    max_epochs=max_epochs,
                    lr_decay_param_reset=args.learning_rate_decay_param_reset,
                    lr_decay_opt_states_reset=args.learning_rate_decay_optimizer_states_reset,
                    decoder=create_checkpoint_decoder(args, exit_stack, context),
                    mxmonitor_pattern=args.monitor_pattern,
                    mxmonitor_stat_func=args.monitor_stat_func,
                    allow_missing_parameters=args.allow_missing_params,
                    existing_parameters=args.params)


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from tempfile import TemporaryDirectory

import mxnet as mx
import numpy as np
import pytest

import sockeye.constants as C
import sockeye.data_io
import sockeye.image_captioning.data_io as data_io
from sockeye import vocab
from sockeye.utils import seedRNGs
from test.common_image_captioning import generate_img_or_feat, tmp_img_captioning_dataset, _FEATURE_SHAPE, _CNN_INPUT_IMAGE_SHAPE

seedRNGs(12)


@pytest.mark.parametrize("source_list, target_sentences, num_samples_per_bucket, expected_source_0, expected_target_0, expected_label_0",
                         [(['1', '2', '3', '4', '100'],
                          [[1, 2, 3], [1, 6, 7], [7, 3], [3, 4, 5, 6], [3, 4]],
                          [2, 2, 1],
                           ['3', '100'], [[ 7., 3.], [ 3., 4.]], [[3., 10.], [4., 10.]])])
def test_raw_list_text_dset_loader(source_list, target_sentences, num_samples_per_bucket,
                                   expected_source_0, expected_target_0, expected_label_0):
    # Test Init object
    buckets = sockeye.data_io.define_parallel_buckets(4, 4, 1, 1.0)
    dset_loader = data_io.RawListTextDatasetLoader(buckets=buckets,
                                       eos_id=10, pad_id=C.PAD_ID)

    assert isinstance(dset_loader, data_io.RawListTextDatasetLoader)
    assert len(dset_loader.buckets)==3

    # Test Load data
    pop_dset_loader = dset_loader.load(source_list, target_sentences, num_samples_per_bucket)

    assert isinstance(pop_dset_loader, sockeye.data_io.ParallelDataSet)
    assert len(pop_dset_loader.source)==3
    assert len(pop_dset_loader.target)==3
    assert len(pop_dset_loader.label)==3
    np.testing.assert_equal(pop_dset_loader.source[0], expected_source_0)
    np.testing.assert_almost_equal(pop_dset_loader.target[0].asnumpy(), expected_target_0)
    np.testing.assert_almost_equal(pop_dset_loader.label[0].asnumpy(), expected_label_0)


@pytest.mark.parametrize("source_list, target_sentences, num_samples_per_bucket",
                         [(['a', 'b', 'c', 'd', 'e'],
                          [[1, 2, 3], [1, 6, 7], [7, 3], [3, 4, 5, 6], [3, 4]],
                           [2, 2, 1])])
def test_image_text_sample_iter(source_list, target_sentences, num_samples_per_bucket):
    batch_size = 2
    image_size = _CNN_INPUT_IMAGE_SHAPE
    buckets = sockeye.data_io.define_parallel_buckets(4, 4, 1, 1.0)
    bucket_batch_sizes = sockeye.data_io.define_bucket_batch_sizes(buckets,
                                                                   batch_size,
                                                                   batch_by_words=False,
                                                                   batch_num_devices=1,
                                                                   data_target_average_len=[None]*len(buckets))
    dset_loader = data_io.RawListTextDatasetLoader(buckets=buckets, eos_id=-1, pad_id=C.PAD_ID)
    with TemporaryDirectory() as work_dir:
        source_list_img = []
        source_list_npy = []
        for s in source_list:
            source_list_img.append(os.path.join(work_dir, s + ".jpg"))
            source_list_npy.append(os.path.join(work_dir, s + ".npy"))
        # Create random images/features
        for s in source_list_img:
            filename = os.path.join(work_dir, s)
            generate_img_or_feat(filename, use_features=False)
        for s in source_list_npy:
            filename = os.path.join(work_dir, s)
            generate_img_or_feat(filename, use_features=True)

        # Test image iterator
        pop_dset_loader = dset_loader.load(source_list_img, target_sentences, num_samples_per_bucket)
        data_iter = data_io.ImageTextSampleIter(pop_dset_loader,
                                                buckets,
                                                batch_size,
                                                bucket_batch_sizes,
                                                image_size,
                                                use_feature_loader=False,
                                                preload_features=False)
        data = data_iter.next()
        assert isinstance(data, mx.io.DataBatch)
        np.testing.assert_equal(data.data[0].asnumpy().shape[1:], image_size)

        # Test iterator feature loader + preload all to memory
        pop_dset_loader = dset_loader.load(source_list_npy, target_sentences, num_samples_per_bucket)
        data_iter = data_io.ImageTextSampleIter(pop_dset_loader,
                                                buckets,
                                                batch_size,
                                                bucket_batch_sizes,
                                                _FEATURE_SHAPE,
                                                use_feature_loader=True,
                                                preload_features=True)
        data = data_iter.next()
        assert isinstance(data, mx.io.DataBatch)
        np.testing.assert_equal(data.data[0].asnumpy().shape[1:], _FEATURE_SHAPE)


def test_get_training_feature_text_data_iters():
    # Test features
    source_list = ['1', '2', '3', '4', '100']
    prefix = "tmp_corpus"
    use_feature_loader = True
    preload_features = True
    train_max_length = 30
    dev_max_length = 30
    expected_mean = 1.0
    expected_std = 1.0
    test_max_length = 30
    batch_size = 5
    if use_feature_loader:
        source_image_size = _FEATURE_SHAPE
    else:
        source_image_size = _CNN_INPUT_IMAGE_SHAPE
    with tmp_img_captioning_dataset(source_list,
                                    prefix,
                                    train_max_length,
                                    dev_max_length,
                                    test_max_length,
                                    use_feature_loader) as data:
        # tmp common vocab
        vcb = vocab.build_from_paths([data['target'], data['target']])

        train_iter, val_iter, config_data, data_info = data_io.get_training_image_text_data_iters(source_root=data['work_dir'],
                                                                                                  source=data['source'],
                                                                                                  target=data['target'],
                                                                                                  validation_source_root=data['work_dir'],
                                                                                                  validation_source=data['validation_source'],
                                                                                                  validation_target=data['validation_target'],
                                                                                                  vocab_target=vcb,
                                                                                                  vocab_target_path=None,
                                                                                                  batch_size=batch_size,
                                                                                                  batch_by_words=False,
                                                                                                  batch_num_devices=1,
                                                                                                  source_image_size=source_image_size,
                                                                                                  fill_up="replicate",
                                                                                                  max_seq_len_target=train_max_length,
                                                                                                  bucketing=True,
                                                                                                  bucket_width=10,
                                                                                                  use_feature_loader=use_feature_loader,
                                                                                                  preload_features=preload_features)
        assert isinstance(train_iter, data_io.ParallelSampleIter)
        assert isinstance(val_iter, data_io.ParallelSampleIter)
        assert isinstance(config_data, data_io.DataConfig)
        assert isinstance(data_info.sources[0], data_io.FileListReader)
        assert data_info.target == data['target']
        assert data_info.source_vocabs is None
        assert data_info.target_vocab is None
        assert config_data.data_statistics.max_observed_len_source == 0
        assert config_data.data_statistics.max_observed_len_target == train_max_length - 1
        assert np.isclose(config_data.data_statistics.length_ratio_mean, expected_mean)
        assert np.isclose(config_data.data_statistics.length_ratio_std, expected_std)

        assert train_iter.batch_size == batch_size
        assert val_iter.batch_size == batch_size
        assert train_iter.default_bucket_key == (0, train_max_length)
        assert val_iter.default_bucket_key == (0, dev_max_length)
        assert train_iter.dtype == 'float32'

        # test some batches
        bos_id = vcb[C.BOS_SYMBOL]
        expected_first_target_symbols = np.full((batch_size,), bos_id, dtype='float32')
        for epoch in range(2):
            while train_iter.iter_next():
                batch = train_iter.next()
                assert len(batch.data) == 2
                assert len(batch.label) == 1
                assert batch.bucket_key in train_iter.buckets
                source = batch.data[0].asnumpy()
                target = batch.data[1].asnumpy()
                label = batch.label[0].asnumpy()
                assert source.shape[0] == target.shape[0] == label.shape[0] == batch_size
                # target first symbol should be BOS
                assert np.array_equal(target[:, 0], expected_first_target_symbols)
                # label first symbol should be 2nd target symbol
                assert np.array_equal(label[:, 0], target[:, 1])
                # each label sequence contains one EOS symbol
                assert np.sum(label == vcb[C.EOS_SYMBOL]) == batch_size
            train_iter.reset()


def test_get_training_image_text_data_iters():
    # Test images
    source_list = ['1', '2', '3', '4', '100']
    prefix = "tmp_corpus"
    use_feature_loader = False
    preload_features = False
    train_max_length = 30
    dev_max_length = 30
    expected_mean = 1.0
    expected_std = 1.0
    test_max_length = 30
    batch_size = 5
    if use_feature_loader:
        source_image_size = _FEATURE_SHAPE
    else:
        source_image_size = _CNN_INPUT_IMAGE_SHAPE
    with tmp_img_captioning_dataset(source_list,
                                    prefix,
                                    train_max_length,
                                    dev_max_length,
                                    test_max_length,
                                    use_feature_loader) as data:
        # tmp common vocab
        vcb = vocab.build_from_paths([data['target'], data['target']])

        train_iter, val_iter, config_data, data_info = data_io.get_training_image_text_data_iters(source_root=data['work_dir'],
                                                                                                  source=data['source'],
                                                                                                  target=data['target'],
                                                                                                  validation_source_root=data['work_dir'],
                                                                                                  validation_source=data['validation_source'],
                                                                                                  validation_target=data['validation_target'],
                                                                                                  vocab_target=vcb,
                                                                                                  vocab_target_path=None,
                                                                                                  batch_size=batch_size,
                                                                                                  batch_by_words=False,
                                                                                                  batch_num_devices=1,
                                                                                                  source_image_size=source_image_size,
                                                                                                  fill_up="replicate",
                                                                                                  max_seq_len_target=train_max_length,
                                                                                                  bucketing=False,
                                                                                                  bucket_width=10,
                                                                                                  use_feature_loader=use_feature_loader,
                                                                                                  preload_features=preload_features)
        assert isinstance(train_iter, data_io.ParallelSampleIter)
        assert isinstance(val_iter, data_io.ParallelSampleIter)
        assert isinstance(config_data, data_io.DataConfig)
        assert isinstance(data_info.sources[0], data_io.FileListReader)
        assert data_info.target == data['target']
        assert data_info.source_vocabs is None
        assert data_info.target_vocab is None
        assert config_data.data_statistics.max_observed_len_source == 0
        assert config_data.data_statistics.max_observed_len_target == train_max_length - 1
        assert np.isclose(config_data.data_statistics.length_ratio_mean, expected_mean)
        assert np.isclose(config_data.data_statistics.length_ratio_std, expected_std)

        assert train_iter.batch_size == batch_size
        assert val_iter.batch_size == batch_size
        assert train_iter.default_bucket_key == (0, train_max_length)
        assert val_iter.default_bucket_key == (0, dev_max_length)
        assert train_iter.dtype == 'float32'

        # test some batches
        bos_id = vcb[C.BOS_SYMBOL]
        expected_first_target_symbols = np.full((batch_size,), bos_id, dtype='float32')
        for epoch in range(2):
            while train_iter.iter_next():
                batch = train_iter.next()
                assert len(batch.data) == 2
                assert len(batch.label) == 1
                assert batch.bucket_key in train_iter.buckets
                source = batch.data[0].asnumpy()
                target = batch.data[1].asnumpy()
                label = batch.label[0].asnumpy()
                assert source.shape[0] == target.shape[0] == label.shape[0] == batch_size
                # target first symbol should be BOS
                assert np.array_equal(target[:, 0], expected_first_target_symbols)
                # label first symbol should be 2nd target symbol
                assert np.array_equal(label[:, 0], target[:, 1])
                # each label sequence contains one EOS symbol
                assert np.sum(label == vcb[C.EOS_SYMBOL]) == batch_size
            train_iter.reset()


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
CLI for extracting image features.
"""
import argparse
import os
import pickle
from contextlib import ExitStack

import mxnet as mx
import numpy as np

from . import arguments
from . import encoder
from . import utils
from .. import constants as C
from ..log import setup_main_logger
from ..translate import _setup_context

# Temporary logger, the real one (logging to a file probably, will be created
# in the main function)
logger = setup_main_logger(__name__, file_logging=False, console=True)


def batching(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def get_pretrained_net(args: argparse.Namespace,
                       context: mx.Context) -> mx.mod.Module:
    # init encoder
    image_cnn_encoder_config = encoder.ImageLoadedCnnEncoderConfig(
        model_path=args.image_encoder_model_path,
        epoch=args.image_encoder_model_epoch,
        layer_name=args.image_encoder_layer,
        encoded_seq_len=0,
        num_embed=100,
        preextracted_features=False
    )  # this num does not matter here

    image_cnn_encoder = encoder.ImageLoadedCnnEncoder(image_cnn_encoder_config)
    symbol = image_cnn_encoder.sym  # this is the net before further encoding
    arg_shapes, out_shapes, aux_shapes = symbol.infer_shape(
        source=(1,) + tuple(args.source_image_size))
    last_layer_shape = out_shapes[-1][1:]

    # Create module
    module = mx.mod.Module(symbol=symbol,
                           data_names=[C.SOURCE_NAME],
                           label_names=None,
                           context=context)
    module.bind(for_training=False, data_shapes=[(C.SOURCE_NAME,
                                                  (args.batch_size,) + tuple(
                                                      args.source_image_size))])

    # Init with pretrained net
    initializers = image_cnn_encoder.get_initializers()
    init = mx.initializer.Mixed(*zip(*initializers))
    module.init_params(init)

    return module, last_layer_shape


def extract_features_forward(im, module, image_root, output_root, batch_size,
                             source_image_size, context):
    batch = mx.nd.zeros((batch_size,) + \
                        tuple(source_image_size), context)
    # Reading
    out_names = []
    for i, v in enumerate(im):
        batch[i] = utils.load_preprocess_image(os.path.join(image_root, v),
                                               source_image_size[1:])
        out_names.append(os.path.join(output_root, v.replace("/", "_")))
    # Forward
    module.forward(mx.io.DataBatch([batch]))
    feats = module.get_outputs()[0].asnumpy()
    # Chunk last batch which might be smaller
    if len(im) < batch_size:
        feats = feats[:len(im)]
    return feats, out_names


def read_list_file(input):
    with open(input, "r") as fd:
        data_list = []
        for i in fd.readlines():
            data_list.append(i.split("\n")[0])
    return data_list


def main():
    params = argparse.ArgumentParser(description='CLI to extract features ' \
                                                 'from images.')
    arguments.add_image_extract_features_cli_args(params)
    args = params.parse_args()

    image_root = os.path.abspath(args.image_root)
    output_root = os.path.abspath(args.output_root)
    output_file = os.path.abspath(args.output)
    size_out_file = os.path.join(output_root, "image_feature_sizes.pkl")
    if os.path.exists(output_root):
        logger.info("Overwriting provided path {}.".format(output_root))
    else:
        os.makedirs(output_root)

    # read image list file
    image_list = read_list_file(args.input)

    # Get pretrained net module (already bind)
    with ExitStack() as exit_stack:
        context = _setup_context(args, exit_stack)
        module, _ = get_pretrained_net(args, context)

        # Extract features
        with open(output_file, "w") as fout:
            for i, im in enumerate(batching(image_list, args.batch_size)):
                logger.info("Processing batch {}/{}".format(i + 1,
                                                            int(np.ceil(len(
                                                                image_list) / args.batch_size))))
                # TODO: enable caching to reuse features and resume computation
                feats, out_names = extract_features_forward(im, module,
                                                            image_root,
                                                            output_root,
                                                            args.batch_size,
                                                            args.source_image_size,
                                                            context)
                # Save to disk
                out_file_names = utils.save_features(out_names, feats)
                # Write to output file
                out_file_names = map(lambda x: os.path.basename(x) + "\n",
                                     out_file_names)
                fout.writelines(out_file_names)

        # Save the image size and feature size
        with open(size_out_file, "wb") as fout:
            pickle.dump({"image_shape": tuple(args.source_image_size),
                         "features_shape": tuple(feats.shape[1:])}, fout)

        # Copy image model to output_folder
        image_encoder_model_path = utils.copy_mx_model_to(
            args.image_encoder_model_path,
            args.image_encoder_model_epoch,
            output_root
        )

        logger.info("Files saved in {}, {} and {}.".format(output_file,
                                                           size_out_file,
                                                           image_encoder_model_path))


if __name__ == "__main__":
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
import os
import pickle
import random
import sys
from contextlib import contextmanager
from tempfile import TemporaryDirectory
from typing import List, Optional, Tuple
from unittest.mock import patch

import mxnet as mx
import numpy as np

import sockeye.average
import sockeye.constants as C
import sockeye.image_captioning.captioner
import sockeye.image_captioning.extract_features
import sockeye.image_captioning.train
from sockeye.evaluate import raw_corpus_bleu, raw_corpus_chrf

try:  # Try to import pillow
    from PIL import Image  # pylint: disable=import-error
except ImportError as e:
    raise RuntimeError("Please install pillow.")

logger = logging.getLogger(__name__)


_DIGITS = "0123456789"
_IMAGE_SHAPE = (100, 100, 3)
_CNN_INPUT_IMAGE_SHAPE = (3, 224, 224)
_FEATURE_SHAPE = (2048, 7, 7)


def generate_img_or_feat(filename, use_features):
    if not use_features:
        imarray = np.random.rand(*_IMAGE_SHAPE) * 255
        im = Image.fromarray(imarray.astype('uint8'))
        im.save(filename)
    else:
        data = np.random.rand(*_FEATURE_SHAPE)
        np.save(filename, data)


def generate_img_text_experiment_files(
                            source_list: List[str],
                            work_dir: str,
                            source_path: str,
                            target_path: str,
                            line_length: int = 9,
                            use_features: bool = False,
                            seed=13):
    random_gen = random.Random(seed)
    with open(source_path, "w") as source_out, open(target_path, "w") as target_out:
        source_list_img = []
        for s in source_list:
            if not use_features:
                filename = s + ".jpg"
            else:
                filename = s + ".npy"
            source_list_img.append(os.path.join(work_dir, filename))
            print(filename, file=source_out)
            digits = [random_gen.choice(_DIGITS) for _ in range(random_gen.randint(1, line_length))]
            print(" ".join(digits), file=target_out)
        # Create random images/features
        for s in source_list_img:
            filename = os.path.join(work_dir, s)
            generate_img_or_feat(filename, use_features)
        # Generate and save the image size and feature size
        size_out_file = os.path.join(work_dir, "image_feature_sizes.pkl")
        with open(size_out_file, "wb") as fout:
            pickle.dump({"image_shape": _CNN_INPUT_IMAGE_SHAPE,
                         "features_shape": _FEATURE_SHAPE}, fout)


@contextmanager
def tmp_img_captioning_dataset(
                        source_list: List[str],
                        prefix: str,
                        train_max_length: int,
                        dev_max_length: int,
                        test_max_length: int,
                        use_features: bool = False,
                        seed_train: int = 13,
                        seed_dev: int = 13):
    with TemporaryDirectory(prefix=prefix) as work_dir:
        # Simple digits files for train/dev data
        train_source_path = os.path.join(work_dir, "train.src")
        train_target_path = os.path.join(work_dir, "train.tgt")
        dev_source_path = os.path.join(work_dir, "dev.src")
        dev_target_path = os.path.join(work_dir, "dev.tgt")
        test_source_path = os.path.join(work_dir, "test.src")
        test_target_path = os.path.join(work_dir, "test.tgt")
        generate_img_text_experiment_files(source_list, work_dir, train_source_path, train_target_path,
                             train_max_length, use_features, seed=seed_train)
        generate_img_text_experiment_files(source_list, work_dir, dev_source_path, dev_target_path,
                                 dev_max_length, use_features, seed=seed_dev)
        generate_img_text_experiment_files(source_list, work_dir, test_source_path, test_target_path,
                                 test_max_length, use_features, seed=seed_dev)
        data = {'work_dir': work_dir,
                'source': train_source_path,
                'target': train_target_path,
                'validation_source': dev_source_path,
                'validation_target': dev_target_path,
                'test_source': test_source_path,
                'test_target': test_target_path}

        yield data


_CAPTION_TRAIN_PARAMS_COMMON = \
    "--use-cpu --max-seq-len {max_len} --source-root {source_root} --source {train_source} --target {train_target}" \
    " --validation-source-root {dev_root} --validation-source {dev_source} --validation-target {dev_target} --output {model} {quiet}" \
    " --seed {seed}"

_CAPTIONER_PARAMS_COMMON = "--use-cpu --models {model}  --source-root {source_root} --input {input} --output {output} {quiet}"

def run_train_captioning(train_params: str,
                        translate_params: str,
                        translate_params_equiv: Optional[str],
                        train_source_path: str,
                        train_target_path: str,
                        dev_source_path: str,
                        dev_target_path: str,
                        test_source_path: str,
                        test_target_path: str,
                        max_seq_len: int = 10,
                        work_dir: Optional[str] = None,
                        seed: int = 13,
                        quiet: bool = False) -> Tuple[float, float, float, float]:
    """
    Train a model and caption a dev set.  Report validation perplexity and BLEU.

    :param train_params: Command line args for model training.
    :param translate_params: First command line args for translation.
    :param translate_params_equiv: Second command line args for captuoning. Should produce the same outputs
    :param train_source_path: Path to the source file.
    :param train_target_path: Path to the target file.
    :param dev_source_path: Path to the development source file.
    :param dev_target_path: Path to the development target file.
    :param test_source_path: Path to the test source file.
    :param test_target_path: Path to the test target file.
    :param max_seq_len: The maximum sequence length.
    :param work_dir: The directory to store the model and other outputs in.
    :param seed: The seed used for training.
    :param quiet: Suppress the console output of training and decoding.
    :return: A tuple containing perplexity, bleu scores for standard and reduced vocab decoding, chrf score.
    """
    source_root = work_dir
    if quiet:
        quiet_arg = "--quiet"
    else:
        quiet_arg = ""
    with TemporaryDirectory(dir=work_dir, prefix="test_train_translate.") as work_dir:
        # Train model
        model_path = os.path.join(work_dir, "model")
        params = "{} {} {}".format(sockeye.image_captioning.train.__file__,
                                   _CAPTION_TRAIN_PARAMS_COMMON.format(
                                       source_root=source_root,
                                       train_source=train_source_path,
                                       train_target=train_target_path,
                                       dev_root=source_root,
                                       dev_source=dev_source_path,
                                       dev_target=dev_target_path,
                                       model=model_path,
                                       max_len=max_seq_len,
                                       seed=seed,
                                       quiet=quiet_arg),
                                   train_params)

        logger.info("Starting training with parameters %s.", train_params)
        with patch.object(sys, "argv", params.split()):
            sockeye.image_captioning.train.main()

        logger.info("Translating with parameters %s.", translate_params)
        # Translate corpus with the 1st params
        out_path = os.path.join(work_dir, "out.txt")
        params = "{} {} {}".format(sockeye.image_captioning.captioner.__file__,
                                   _CAPTIONER_PARAMS_COMMON.format(model=model_path,
                                                                   source_root=source_root,
                                                                   input=test_source_path,
                                                                   output=out_path,
                                                                   quiet=quiet_arg),
                                   translate_params)
        with patch.object(sys, "argv", params.split()):
            sockeye.image_captioning.captioner.main()

        # Translate corpus with the 2nd params
        if translate_params_equiv is not None:
            out_path_equiv = os.path.join(work_dir, "out_equiv.txt")
            params = "{} {} {}".format(sockeye.image_captioning.captioner.__file__,
                                   _CAPTIONER_PARAMS_COMMON.format(model=model_path,
                                                                   source_root=source_root,
                                                                   input=test_source_path,
                                                                   output=out_path_equiv,
                                                                   quiet=quiet_arg),
                                    translate_params_equiv)
            with patch.object(sys, "argv", params.split()):
                sockeye.image_captioning.captioner.main()
            # read-in both outputs, ensure they are the same
            with open(out_path, 'rt') as f:
                lines = f.readlines()
            with open(out_path_equiv, 'rt') as f:
                lines_equiv = f.readlines()
            assert all(a == b for a, b in zip(lines, lines_equiv))

        # test averaging
        points = sockeye.average.find_checkpoints(model_path=model_path,
                                                  size=1,
                                                  strategy='best',
                                                  metric=C.PERPLEXITY)
        assert len(points) > 0
        averaged_params = sockeye.average.average(points)
        assert averaged_params

        # get best validation perplexity
        metrics = sockeye.utils.read_metrics_file(path=os.path.join(model_path, C.METRICS_NAME))
        perplexity = min(m[C.PERPLEXITY + '-val'] for m in metrics)
        hypotheses = open(out_path, "r").readlines()
        references = open(test_target_path, "r").readlines()
        assert len(hypotheses) == len(references)
        # compute metrics
        bleu = raw_corpus_bleu(hypotheses=hypotheses, references=references, offset=0.01)
        chrf = raw_corpus_chrf(hypotheses=hypotheses, references=references)

        return perplexity, bleu, chrf


_EXTRACT_FEATURES_PARAMS_COMMON = \
    "--use-cpu --image-root {image_root} --input {source_file} --output-root {output_root} " \
    "--output {output_file} --image-encoder-model-path {image_encoder_model_path}"


def run_extract_features_captioning(source_image_size: tuple,
                                    batch_size: int,
                                    extract_params: str,
                                    source_files: List[str],
                                    image_root: str) -> None:

    with TemporaryDirectory(dir=image_root, prefix="test_extract_feats") as work_dir:
        model_path = os.path.join(work_dir, '2-conv-layer')
        epoch = 0
        # Create net and save to disk
        create_simple_and_save_to_disk(model_path, epoch, source_image_size, batch_size)

        # Extract features
        for s in source_files:
            with TemporaryDirectory(dir=work_dir, prefix="extracted_feats") as local_work_dir:
                output_root = local_work_dir
                output_file = os.path.join(local_work_dir, "random.features")
                params = "{} {} {}".format(sockeye.image_captioning.extract_features.__file__,
                                           _EXTRACT_FEATURES_PARAMS_COMMON.format(
                                               image_root=image_root,
                                               source_file=s,
                                               output_root=output_root,
                                               output_file=output_file,
                                               image_encoder_model_path=model_path
                                           ),
                                           extract_params)

                logger.info("Starting feature extractopm with parameters %s.", extract_params)
                with patch.object(sys, "argv", params.split()):
                    sockeye.image_captioning.extract_features.main()


def create_simple_and_save_to_disk(prefix, iteration, source_image_size, batch_size):
    # init model
    sym = get_2convnet_symbol()
    mod = mx.mod.Module(sym)
    mod.bind(data_shapes=[('data', (batch_size,) + source_image_size)],
             label_shapes=[('softmax_label', (batch_size, 1))])
    mod.init_params()
    # save
    mod.save_checkpoint(prefix, iteration)


def get_2convnet_symbol():
    data = mx.symbol.Variable('data')
    # first conv
    conv1 = mx.symbol.Convolution(data=data, kernel=(5,5), num_filter=20, name='conv1')
    tanh1 = mx.symbol.Activation(data=conv1, act_type="tanh")
    pool1 = mx.symbol.Pooling(data=tanh1, pool_type="max",
                              kernel=(2,2), stride=(2,2))
    # second conv
    conv2 = mx.symbol.Convolution(data=pool1, kernel=(5,5), num_filter=50, name='conv2')
    tanh2 = mx.symbol.Activation(data=conv2, act_type="tanh")
    pool2 = mx.symbol.Pooling(data=tanh2, pool_type="max",
                              kernel=(2,2), stride=(2,2))
    flatten = mx.symbol.Flatten(data=pool2)
    fc2 = mx.symbol.FullyConnected(data=flatten, num_hidden=1)
    # loss
    outsym = mx.symbol.SoftmaxOutput(data=fc2, name='softmax')
    return outsym


# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import copy
import logging
from operator import attrgetter
from typing import List, Optional, Tuple, Set

import mxnet as mx
import numpy as np

logger = logging.getLogger(__name__)

# Represents a list of raw constraints for a sentence. Each constraint is a list of target-word IDs.
RawConstraintList = List[List[int]]


class ConstrainedHypothesis:
    """
    Represents a set of words and phrases that must appear in the output.
    A constraint is of two types: sequence or non-sequence.
    A non-sequence constraint is a single word and can therefore be followed by anything,
    whereas a sequence constraint must be followed by a particular word (the next word in the sequence).
    This class also records which constraints have been met.

    A list of raw constraints is maintained internally as two parallel arrays. The following raw constraint
    represents two phrases that must appear in the output: 14 and 19 35 14.

        raw constraint: [[14], [19, 35, 14]]

    This is represented internally as:

        constraints: [14 19 35 14]
        is_sequence: [ 1  1  0  0

    :param constraint_list: A list of zero or raw constraints (each represented as a list of integers).
    :param eos_id: The end-of-sentence ID.
    """

    def __init__(self,
                 constraint_list: RawConstraintList,
                 eos_id: int) -> None:

        # `constraints` records the words of the constraints, as a list (duplicates allowed).
        # `is_sequence` is a parallel array that records, for each corresponding constraint,
        #    whether the current word is the non-final word of a phrasal constraint.
        self.constraints = []  # type: List[int]
        self.is_sequence = []  # type: List[bool]
        for phrase in constraint_list:
            self.constraints += phrase
            self.is_sequence += [True] * len(phrase)
            self.is_sequence[-1] = False

        self.eos_id = eos_id

        # no constraints have been met
        self.met = [False for x in self.constraints]
        self.last_met = -1

    def __len__(self) -> int:
        """
        :return: The number of constraints.
        """
        return len(self.constraints)

    def __str__(self) -> str:
        s = []
        for i, word_id in enumerate(self.constraints):
            s.append(str(word_id) if self.met[i] is False else 'X')
            if self.is_sequence[i]:
                s.append('->')
        return ' '.join(s)

    def size(self) -> int:
        """
        :return: the number of constraints
        """
        return len(self.constraints)

    def num_met(self) -> int:
        """
        :return: the number of constraints that have been met.
        """
        return sum(self.met)

    def num_needed(self) -> int:
        """
        :return: the number of un-met constraints.
        """
        return self.size() - self.num_met()

    def allowed(self) -> Set[int]:
        """
        Returns the set of constrained words that could follow this one.
        For unfinished phrasal constraints, it is the next word in the phrase.
        In other cases, it is the list of all unmet constraints.
        If all constraints are met, an empty set is returned.

        :return: The ID of the next required word, or -1 if any word can follow
        """
        items = set()  # type: Set[int]
        # Add extensions of a started-but-incomplete sequential constraint
        if self.last_met != -1 and self.is_sequence[self.last_met] == 1:
            word_id = self.constraints[self.last_met + 1]
            if word_id != self.eos_id or self.num_needed() == 1:
                items.add(word_id)

        # Add all constraints that aren't non-initial sequences
        else:
            for i, word_id in enumerate(self.constraints):
                if not self.met[i] and (i == 0 or not self.is_sequence[i - 1]):
                    if word_id != self.eos_id or self.num_needed() == 1:
                        items.add(word_id)

        return items

    def finished(self) -> bool:
        """
        Return true if all the constraints have been met.

        :return: True if all the constraints are met.
        """
        return self.num_needed() == 0

    def is_valid(self, wordid) -> bool:
        """
        Ensures </s> is only generated when the hypothesis is completed.

        :param wordid: The wordid to validate.
        :return: True if all constraints are already met or the word ID is not the EOS id.
        """
        return self.finished() or wordid != self.eos_id or (self.num_needed() == 1 and self.eos_id in self.allowed())

    def advance(self, word_id: int) -> 'ConstrainedHypothesis':
        """
        Updates the constraints object based on advancing on word_id.
        There is a complication, in that we may have started but not
        yet completed a multi-word constraint.  We need to allow constraints
        to be added as unconstrained words, so if the next word is
        invalid, we must "back out" of the current (incomplete) phrase,
        re-setting all of its words as unmet.

        :param word_id: The word ID to advance on.
        :return: A deep copy of the object, advanced on word_id.
        """

        obj = copy.deepcopy(self)

        # First, check if we're updating a sequential constraint.
        if obj.last_met != -1 and obj.is_sequence[obj.last_met] == 1:
            if word_id == obj.constraints[obj.last_met + 1]:
                # Here, the word matches what we expect next in the constraint, so we update everything
                obj.met[obj.last_met + 1] = True
                obj.last_met += 1
            else:
                # Here, the word is not the expected next word of the constraint, so we back out of the constraint.
                index = obj.last_met
                while obj.is_sequence[index]:
                    obj.met[index] = False
                    index -= 1
                obj.last_met = -1

        # If not, check whether we're meeting a single-word constraint
        else:
            # Build a list from all constraints of tuples of the
            # form (constraint, whether it's a non-initial sequential, whether it's been met)
            constraint_tuples = list(zip(obj.constraints, [False] + obj.is_sequence[:-1], obj.met))
            # We are searching for an unmet constraint (word_id) that is not the middle of a phrase and is not met
            query = (word_id, False, False)
            try:
                pos = constraint_tuples.index(query)
                obj.met[pos] = True
                obj.last_met = pos
            except ValueError:
                # query not found; identical but duplicated object will be returned
                pass

        return obj


def init_batch(raw_constraints: List[Optional[RawConstraintList]],
               beam_size: int,
               start_id: int,
               eos_id: int) -> List[Optional[ConstrainedHypothesis]]:
    """
    :param raw_constraints: The list of raw constraints (list of list of IDs).
    :param beam_size: The beam size.
    :param start_id: The target-language vocabulary ID of the SOS symbol.
    :param eos_id: The target-language vocabulary ID of the EOS symbol.
    :return: A list of ConstrainedHypothesis objects (shape: (batch_size * beam_size,)).
    """
    constraints = [None] * (len(raw_constraints) * beam_size)  # type: List[Optional[ConstrainedHypothesis]]
    if any(raw_constraints):
        for i, raw_list in enumerate(raw_constraints):
            num_constraints = sum([len(phrase) for phrase in raw_list]) if raw_list is not None else 0
            if num_constraints > 0:
                hyp = ConstrainedHypothesis(raw_list, eos_id)
                idx = i * beam_size
                constraints[idx:idx + beam_size] = [hyp.advance(start_id) for x in range(beam_size)]

    return constraints


def get_bank_sizes(num_constraints: int,
                   beam_size: int,
                   candidate_counts: List[int]) -> List[int]:
    """
    Evenly distributes the beam across the banks, where each bank is a portion of the beam devoted
    to hypotheses having met the same number of constraints, 0..num_constraints.
    After the assignment, banks with more slots than candidates are adjusted.

    :param num_constraints: The number of constraints.
    :param beam_size: The beam size.
    :param candidate_counts: The empirical counts of number of candidates in each bank.
    :return: A distribution over banks.
    """

    num_banks = num_constraints + 1
    bank_size = beam_size // num_banks
    remainder = beam_size - bank_size * num_banks

    # Distribute any remainder to the end
    assigned = [bank_size for x in range(num_banks)]
    assigned[-1] += remainder

    # Now, moving right to left, push extra allocation to earlier buckets.
    # This encodes a bias for higher buckets, but if no candidates are found, space
    # will be made in lower buckets. This may not be the best strategy, but it is important
    # that you start pushing from the bucket that is assigned the remainder, for cases where
    # num_constraints >= beam_size.
    for i in reversed(range(num_banks)):
        overfill = assigned[i] - candidate_counts[i]
        if overfill > 0:
            assigned[i] -= overfill
            assigned[(i - 1) % num_banks] += overfill

    return assigned


class ConstrainedCandidate:
    """
    Object used to hold candidates for the beam in topk().

    :param row: The row in the scores matrix.
    :param col: The column (word ID) in the scores matrix.
    :param score: the associated accumulated score.
    :param hypothesis: The ConstrainedHypothesis containing information about met constraints.
    """

    __slots__ = ('row', 'col', 'score', 'hypothesis')

    def __init__(self,
                 row: int,
                 col: int,
                 score: float,
                 hypothesis: ConstrainedHypothesis) -> None:
        self.row = row
        self.col = col
        self.score = score
        self.hypothesis = hypothesis

    def __hash__(self):
        return hash((self.row, self.col))

    def __eq__(self, other):
        return self.row == other.row and self.col == other.col

    def __str__(self):
        return '({}, {}, {}, {})'.format(self.row, self.col, self.score, self.hypothesis.num_met())


def topk(batch_size: int,
         beam_size: int,
         inactive: mx.ndarray,
         scores: mx.ndarray,
         hypotheses: List[ConstrainedHypothesis],
         best_ids: mx.ndarray,
         best_word_ids: mx.ndarray,
         seq_scores: mx.ndarray,
         context: mx.context.Context) -> Tuple[np.array, np.array, np.array, List[ConstrainedHypothesis], mx.nd.array]:
    """
    Builds a new topk list such that the beam contains hypotheses having completed different numbers of constraints.
    These items are built from three different types: (1) the best items across the whole
    scores matrix, (2) the set of words that must follow existing constraints, and (3) k-best items from each row.

    :param batch_size: The number of segments in the batch.
    :param beam_size: The length of the beam for each segment.
    :param inactive: Array listing inactive rows (shape: (beam_size,)).
    :param scores: The scores array (shape: (beam_size, target_vocab_size)).
    :param hypotheses: The list of hypothesis objects.
    :param best_ids: The current list of best hypotheses (shape: (beam_size,)).
    :param best_word_ids: The parallel list of best word IDs (shape: (beam_size,)).
    :param seq_scores: (shape: (beam_size, 1)).
    :param context: The MXNet device context.
    :return: A tuple containing the best hypothesis rows, the best hypothesis words, the scores,
        the updated constrained hypotheses, and the updated set of inactive hypotheses.
    """

    for sentno in range(batch_size):
        rows = slice(sentno * beam_size, (sentno + 1) * beam_size)
        if hypotheses[rows.start] is not None and hypotheses[rows.start].size() > 0:
            best_ids[rows], best_word_ids[rows], seq_scores[rows], \
            hypotheses[rows], inactive[rows] = _topk(beam_size,
                                                     inactive[rows],
                                                     scores[rows],
                                                     hypotheses[rows],
                                                     best_ids[rows] - rows.start,
                                                     best_word_ids[rows],
                                                     seq_scores[rows],
                                                     context)

            # offsetting since the returned smallest_k() indices were slice-relative
            best_ids[rows] += rows.start
        else:
            # If there are no constraints for this sentence in the batch, everything stays
            # the same, except we need to mark all hypotheses as active
            inactive[rows] = 0

    return (best_ids, best_word_ids, seq_scores, hypotheses, inactive)


def _topk(beam_size: int,
          inactive: mx.ndarray,
          scores: mx.ndarray,
          hypotheses: List[ConstrainedHypothesis],
          best_ids: mx.ndarray,
          best_word_ids: mx.ndarray,
          sequence_scores: mx.ndarray,
          context: mx.context.Context) -> Tuple[np.array, np.array, np.array, List[ConstrainedHypothesis], mx.nd.array]:
    """
    Builds a new topk list such that the beam contains hypotheses having completed different numbers of constraints.
    These items are built from three different types: (1) the best items across the whole
    scores matrix, (2) the set of words that must follow existing constraints, and (3) k-best items from each row.

    :param beam_size: The length of the beam for each segment.
    :param inactive: Array listing inactive rows (shape: (beam_size,)).
    :param scores: The scores array (shape: (beam_size, target_vocab_size)).
    :param hypotheses: The list of hypothesis objects.
    :param best_ids: The current list of best hypotheses (shape: (beam_size,)).
    :param best_word_ids: The parallel list of best word IDs (shape: (beam_size,)).
    :param sequence_scores: (shape: (beam_size, 1)).
    :param context: The MXNet device context.
    :return: A tuple containing the best hypothesis rows, the best hypothesis words, the scores,
        the updated constrained hypotheses, and the updated set of inactive hypotheses.
    """

    num_constraints = hypotheses[0].size()

    candidates = set()
    # (1) Add all of the top-k items (which were passed) in as long as they pass the constraints
    for row, col, seq_score in zip(best_ids, best_word_ids, sequence_scores):
        row = int(row.asscalar())
        col = int(col.asscalar())
        if hypotheses[row].is_valid(col):
            seq_score = float(seq_score.asscalar())
            new_item = hypotheses[row].advance(col)
            cand = ConstrainedCandidate(row, col, seq_score, new_item)
            candidates.add(cand)

    # For each hypothesis, we add (2) all the constraints that could follow it and
    # (3) the best item (constrained or not) in that row
    best_next = mx.ndarray.argmin(scores, axis=1)
    for row in range(beam_size):
        if inactive[row]:
            continue

        hyp = hypotheses[row]

        # (2) add all the constraints that could extend this
        nextones = hyp.allowed()

        # (3) add the single-best item after this (if it's valid)
        col = int(best_next[row].asscalar())
        if hyp.is_valid(col):
            nextones.add(col)

        # Now, create new candidates for each of these items
        for col in nextones:
            new_item = hyp.advance(col)
            score = scores[row, col].asscalar()
            cand = ConstrainedCandidate(row, col, score, new_item)
            candidates.add(cand)

    # Sort the candidates. After allocating the beam across the banks, we will pick the top items
    # for each bank from this list
    sorted_candidates = sorted(candidates, key=attrgetter('score'))

    # The number of hypotheses in each bank
    counts = [0 for x in range(num_constraints + 1)]
    for cand in sorted_candidates:
        counts[cand.hypothesis.num_met()] += 1

    # Adjust allocated bank sizes if there are too few candidates in any of them
    bank_sizes = get_bank_sizes(num_constraints, beam_size, counts)

    # Sort the candidates into the allocated banks
    pruned_candidates = []  # type: List[ConstrainedCandidate]
    for i, cand in enumerate(sorted_candidates):
        bank = cand.hypothesis.num_met()

        if bank_sizes[bank] > 0:
            pruned_candidates.append(cand)
            bank_sizes[bank] -= 1

    inactive[:len(pruned_candidates)] = 0

    # Pad the beam so array assignment still works
    if len(pruned_candidates) < beam_size:
        inactive[len(pruned_candidates):] = 1
        pruned_candidates += [pruned_candidates[len(pruned_candidates) - 1]] * (beam_size - len(pruned_candidates))

    return (np.array([x.row for x in pruned_candidates]),
            np.array([x.col for x in pruned_candidates]),
            np.array([[x.score] for x in pruned_candidates]),
            [x.hypothesis for x in pruned_candidates],
            inactive)


def main():
    """
    Usage: python3 -m sockeye.lexical_constraints [--bpe BPE_MODEL]

    Reads sentences and constraints on STDIN (tab-delimited) and generates the JSON format
    that can be used when passing `--json-input` to sockeye.translate.

    e.g.,

        echo -e "Dies ist ein Test .\tThis is\ttest" | python3 -m sockeye.lexical_constraints

    will produce the following JSON object:

        { "text": "Dies ist ein Test .", "constraints": ["This is", "test"] }

    Make sure you apply all preprocessing (tokenization, BPE, etc.) to both the source and the target-side constraints.
    You can then translate this object by passing it to Sockeye on STDIN as follows:

        python3 -m sockeye.translate -m /path/to/model --json-input --beam-size 20 --beam-prune 20

    (Note the recommended Sockeye parameters).
    """
    import sys
    import json

    for line in sys.stdin:
        line = line.rstrip()

        # Constraints are in fields 2+
        source, *constraints = line.split('\t')

        obj = {'text': source}
        if len(constraints) > 0:
            obj['constraints'] = constraints

        print(json.dumps(obj, ensure_ascii=False), flush=True)


if __name__ == '__main__':
    main()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Visualize the checkpoints of the model: image, ground truth caption and
predicted caption.
"""
import argparse
import os

try:  # Try to import pillow
    from PIL import Image  # pylint: disable=import-error
except ImportError as e:
    raise RuntimeError("Please install pillow.")

try:  # Try to import matplotlib
    import matplotlib  # pylint: disable=import-error
except ImportError as e:
    raise RuntimeError("Please install matplotlib.")
matplotlib.use('Agg')
import matplotlib.pyplot as plt


def format_text_for_visualization(c, n):
    c = c.split(" ")
    c[0] = c[0].title()
    out = ""
    for j in range(0, len(c)):
        out += c[j]
        if j == len(c)-1:
            out += "."
        else:
            if (j + 1) % n == 0:
                out += "\n"
            else:
                out += " "
    return out

def main():
    params = argparse.ArgumentParser(
        description="CLI to visualize the captions along with images and "
                    "ground truth."
    )
    params.add_argument("-d", "--image-root",
                        help="Absolute path of the dataset root where the "
                             "images are stored.")
    params.add_argument("-i", "--source",
                        help="File containing the images or features used to "
                             "generate the captions.")
    params.add_argument("-c", "--prediction",
                        help="File containing the captions. Each line "
                             "corresponds to a line in the source.")
    params.add_argument("-a", "--ground-truth",
                        default=None,
                        help="File file containing the ground-truth captions "
                             "(optional).")
    params.add_argument("-s", "--save-to-folder",
                        default=None,
                        help="Folder to save the visualizations.")
    params.add_argument("-si", "--skip-images",
                        default=2,
                        help="Number of images to skip for visualization.")
    params.add_argument("-nc", "--number-of-columns",
                        default=4,
                        help="Number of columns in the subplot (better if even "
                             "number).")
    args = params.parse_args()

    skip = args.skip_images
    N = M = args.number_of_columns

    # adjust this if visualization is not nice
    len_newline = 9
    fontsize = 10
    figsize = (30, 20)

    # Collect results in a better data structure (dict)
    # * Read predictions and image dir
    fs = open(args.source)
    fc = open(args.prediction)
    predictions = {}
    for s, c in zip(fs.readlines(), fc.readlines()):
        predictions[s] = c  # just keep one sentence
    fs.close()
    fc.close()
    # * Read ground truth optionally
    ground_truth = {}
    if args.ground_truth is not None:
        fgt = open(args.ground_truth)
        fs = open(args.source)
        for s, gt in zip(fs.readlines(), fgt.readlines()):
            if s in ground_truth:
                ground_truth[s].append(gt)
            else:
                ground_truth[s] = [gt]
        fgt.close()
    fs.close()

    # Prepare output folder, if needed
    if args.save_to_folder is not None:
        fontsize = 15
        if not os.path.exists(args.save_to_folder):
            os.makedirs(args.save_to_folder)

    # Visualization
    plt.ioff()
    fig, axs = plt.subplots(N, M, figsize=figsize)
    fig.tight_layout()
    i = 0
    ii = 1
    for s in predictions.keys():  # Go over images (dict[image]=caption)
        if ii%skip==0: # maybe you do not want to display all images
            c = predictions[s]
            if len(ground_truth)>0:
                gts = ground_truth[s] # list
            s = s.split("\n")[0]
            c = c.split("\n")[0]
            # Display image
            image = Image.open(os.path.join(args.image_root, s))
            if 'RGB' not in image.mode:
                axs[i//N%M, i%N].imshow(image, cmap='gray')
            else:
                axs[i//N%M, i%N].imshow(image)
            # Display predicted caption
            axs[i//N%M, i%N].axis("off")
            axs[(i+1)//N%M, (i+1)%N].text(0, 0.9,
                                  format_text_for_visualization(c, len_newline),
                                  fontsize=fontsize,
                                  bbox={'facecolor': 'white',
                                        'alpha': 0.85,
                                        'pad': 2})
            # Display ground-truth caption(s) optionally
            if len(ground_truth)>0:
                gt_vis = ""
                for j, gt in enumerate(gts):
                    gt = gt.split("\n")[0]
                    gt_vis += \
                        "* " + format_text_for_visualization(gt, len_newline) \
                        + "\n"
                axs[(i+1)//N%M, (i+1)%N].text(0, 0, gt_vis,
                                              fontsize=fontsize,
                                              bbox={'facecolor': 'green',
                                                    'alpha': 0.3,
                                                    'pad': 2})
            axs[(i+1)//N%M, (i+1)%N].axis("off")
            i += 2

            # Show or save to disk
            if i % (N * M) == 0:
                if args.save_to_folder is None:
                    plt.show()
                else:
                    plt.savefig(os.path.join(args.save_to_folder,
                                             str(ii).zfill(6) + '.png'),
                                bbox_inches='tight')
                i = 0
                # Reset axes, clean up
                for k in range(N):
                    for j in range(M):
                        axs[k, j].cla()
                        axs[k, j].axis("off")
        ii += 1


if __name__ == "__main__":
    main()





# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
Implements data iterators and I/O related functions for image-to-sequence
models.
"""
import logging
import time
from typing import Any, Dict, Iterable, List, Optional, Tuple

import mxnet as mx
import numpy as np

from .utils import load_features, load_feature, load_preprocess_images
from .. import constants as C
from .. import vocab
from ..data_io import ParallelDataSet, RawParallelDatasetLoader, \
    BucketBatchSize, FileListReader, SequenceReader, DataConfig, DataInfo, \
    ParallelSampleIter
from ..data_io import get_target_bucket, get_data_statistics, \
    define_empty_source_parallel_buckets, define_bucket_batch_sizes

logger = logging.getLogger(__name__)


class RawListTextDatasetLoader:
    """
    Loads a data set of variable-length parallel list of string and target sequences into buckets of NDArrays.
    The list of strings are not converted to NDArrays, because we assume that the dataset does not fit in memory.
    We assume that the used data iterator knows how to load the data from disk to memory every time a batch is consumed.
    Note: it does not support multiple source, like `sockeye.data_io.RawParallelDatasetLoader`.

    :param buckets: Bucket list.
    :param eos_id: End-of-sentence id.
    :param pad_id: Padding id.
    :param eos_id: Unknown id.
    :param dtype: Data type.
    """

    def __init__(self,
                 buckets: List[Tuple[int, int]],
                 eos_id: int,
                 pad_id: int,
                 dtype: str = 'float32') -> None:
        self.buckets = buckets
        self.eos_id = eos_id
        self.pad_id = pad_id
        self.dtype = dtype

    def load(self,
             source_list: Iterable[List[str]],
             target_sentences: Iterable[List[Any]],
             num_samples_per_bucket: List[int]) -> 'ParallelDataSet':
        """
        Creates a parallel dataset base on source list of strings and target sentences.
        Returns a `sockeye.data_io.ParallelDataSet`.

        :param source_list: Source list of strings (e.g., filenames).
        :param target_sentences: Target sentences used to do bucketing.
        :param num_samples_per_bucket: Number of samples per bucket.
        :return: Returns a parallel dataset `sockeye.data_io.ParallelDataSet`.
        """
        assert len(num_samples_per_bucket) == len(self.buckets)

        data_source = [np.full((num_samples, ), self.pad_id, dtype=object)
                       for num_samples in num_samples_per_bucket]
        # data_source is a List[numpy.array[str]] which semantic is bucket, index, str
        # Its loading to memory is deferred to the iterator, since the full data
        # is supposed to not fit in memory.
        data_target = [np.full((num_samples, target_len), self.pad_id, dtype=self.dtype)
                       for (source_len, target_len), num_samples in zip(self.buckets, num_samples_per_bucket)]
        data_label = [np.full((num_samples, target_len), self.pad_id, dtype=self.dtype)
                      for (source_len, target_len), num_samples in zip(self.buckets, num_samples_per_bucket)]

        bucket_sample_index = [0 for buck in self.buckets]

        # track amount of padding introduced through bucketing
        num_tokens_target = 0
        num_pad_target = 0

        # Bucket sentences as padded np arrays
        for source, target in zip(source_list, target_sentences):
            target_len = len(target)
            buck_index, buck = get_target_bucket(self.buckets, target_len)
            if buck is None:
                continue  # skip this sentence pair

            num_tokens_target += buck[1]
            num_pad_target += buck[1] - target_len

            sample_index = bucket_sample_index[buck_index]
            data_source[buck_index][sample_index] = source
            data_target[buck_index][sample_index, :target_len] = target
            # NOTE(fhieber): while this is wasteful w.r.t memory, we need to explicitly create the label sequence
            # with the EOS symbol here sentence-wise and not per-batch due to variable sequence length within a batch.
            # Once MXNet allows item assignments given a list of indices (probably MXNet 1.0): e.g a[[0,1,5,2]] = x,
            # we can try again to compute the label sequence on the fly in next().
            data_label[buck_index][sample_index, :target_len] = target[1:] + [self.eos_id]

            bucket_sample_index[buck_index] += 1

        for i in range(len(data_source)):
            data_target[i] = mx.nd.array(data_target[i], dtype=self.dtype)
            data_label[i] = mx.nd.array(data_label[i], dtype=self.dtype)

        if num_tokens_target > 0:
            logger.info("Created bucketed parallel data set. Introduced padding: target=%.1f%%)",
                        num_pad_target / num_tokens_target * 100)

        return ParallelDataSet(data_source, data_target, data_label)


def get_validation_image_text_data_iter(data_loader: RawParallelDatasetLoader,
                             validation_source_root:str,
                             validation_source: str,
                             validation_target: str,
                             buckets: List[Tuple[int, int]],
                             bucket_batch_sizes: List[BucketBatchSize],
                             source_image_size: tuple,
                             vocab_target: vocab.Vocab,
                             max_seq_len_target: int,
                             batch_size: int,
                             fill_up: str,
                             use_feature_loader: bool = False,
                             preload_features: bool = False) -> 'ParallelSampleIter':
    """
    Returns a ParallelSampleIter for the validation data.
    """
    logger.info("=================================")
    logger.info("Creating validation data iterator")
    logger.info("=================================")

    validation_source_images = [FileListReader(validation_source, validation_source_root)]
    validation_target_sentences = SequenceReader(validation_target, vocab_target, add_bos=True, limit=None)

    validation_data_statistics = get_data_statistics(source_readers=None,
                                          target_reader=validation_target_sentences,
                                          buckets=buckets,
                                          length_ratio_mean=1.0,
                                          length_ratio_std=1.0,
                                          source_vocabs=[None],
                                          target_vocab=vocab_target)
    validation_data_statistics.log(bucket_batch_sizes)

    validation_data = data_loader.load(validation_source_images[0],
                                       validation_target_sentences,
                                       validation_data_statistics.num_sents_per_bucket).fill_up(bucket_batch_sizes,
                                                                                                fill_up)
    return ImageTextSampleIter(data=validation_data,
                               buckets=buckets,
                               batch_size=batch_size,
                               bucket_batch_sizes=bucket_batch_sizes,
                               image_size=source_image_size,
                               use_feature_loader=use_feature_loader,
                               preload_features=preload_features)


def get_training_image_text_data_iters(source_root: str,
                            source: str, target: str,
                            validation_source_root: str,
                            validation_source: str, validation_target: str,
                            vocab_target: vocab.Vocab,
                            vocab_target_path: Optional[str],
                            batch_size: int,
                            batch_by_words: bool,
                            batch_num_devices: int,
                            source_image_size: tuple,
                            fill_up: str,
                            max_seq_len_target: int,
                            bucketing: bool,
                            bucket_width: int,
                            use_feature_loader: bool = False,
                            preload_features: bool = False) -> Tuple['ParallelSampleIter',
                                                                           'ParallelSampleIter',
                                                                           'DataConfig', 'DataInfo']:
    """
    Returns data iterators for training and validation data.

    :param source_root: Path to source images since the file in source contains relative paths.
    :param source: Path to source training data.
    :param target: Path to target training data.
    :param validation_source_root: Path to validation source images since the file in validation_source contains relative paths.
    :param validation_source: Path to source validation data.
    :param validation_target: Path to target validation data.
    :param vocab_target: Target vocabulary.
    :param vocab_target_path: Path to target vocabulary.
    :param batch_size: Batch size.
    :param batch_by_words: Size batches by words rather than sentences.
    :param batch_num_devices: Number of devices batches will be parallelized across.
    :param source_image_size: size to resize the image to (for iterator)
    :param fill_up: Fill-up strategy for buckets.
    :param max_seq_len_target: Maximum target sequence length.
    :param bucketing: Whether to use bucketing.
    :param bucket_width: Size of buckets.
    :param use_feature_loader: If True, features are loaded instead of images.
    :param preload_features: If use_feature_loader si True, this enables load all the feature to memory
    :return: Tuple of (training data iterator, validation data iterator, data config).
    """
    logger.info("===============================")
    logger.info("Creating training data iterator")
    logger.info("===============================")

    # define buckets
    buckets = define_empty_source_parallel_buckets(max_seq_len_target, bucket_width) if bucketing else [
        (0, max_seq_len_target)]

    source_images = [FileListReader(source, source_root)]
    target_sentences = SequenceReader(target, vocab_target, add_bos=True)

    # 2. pass: Get data statistics only on target (source not considered)
    data_statistics = get_data_statistics(source_readers=None,
                                          target_reader=target_sentences,
                                          buckets=buckets,
                                          length_ratio_mean=1.0,
                                          length_ratio_std=1.0,
                                          source_vocabs=[None],
                                          target_vocab=vocab_target)

    bucket_batch_sizes = define_bucket_batch_sizes(buckets,
                                                   batch_size,
                                                   batch_by_words,
                                                   batch_num_devices,
                                                   data_statistics.average_len_target_per_bucket)

    data_statistics.log(bucket_batch_sizes)

    data_loader = RawListTextDatasetLoader(buckets=buckets,
                                           eos_id=vocab_target[C.EOS_SYMBOL],
                                           pad_id=C.PAD_ID)

    training_data = data_loader.load(source_images[0], target_sentences,
                                     data_statistics.num_sents_per_bucket).fill_up(bucket_batch_sizes, fill_up)

    data_info = DataInfo(sources=source_images,
                         target=target,
                         source_vocabs=None,
                         target_vocab=vocab_target_path,
                         shared_vocab=False,
                         num_shards=1)

    config_data = DataConfig(data_statistics=data_statistics,
                             max_seq_len_source=0,
                             max_seq_len_target=max_seq_len_target,
                             num_source_factors=len(source_images))

    # Add useful stuff to config_data
    config_data.source_root = source_root
    config_data.validation_source_root = validation_source_root
    config_data.use_feature_loader = use_feature_loader

    train_iter = ImageTextSampleIter(data=training_data,
                                     buckets=buckets,
                                     batch_size=batch_size,
                                     bucket_batch_sizes=bucket_batch_sizes,
                                     image_size=source_image_size,
                                     use_feature_loader=use_feature_loader,
                                     preload_features=preload_features)

    validation_iter = get_validation_image_text_data_iter(data_loader=data_loader,
                                               validation_source_root=validation_source_root,
                                               validation_source=validation_source,
                                               validation_target=validation_target,
                                               buckets=buckets,
                                               bucket_batch_sizes=bucket_batch_sizes,
                                               source_image_size=source_image_size,
                                               vocab_target=vocab_target,
                                               max_seq_len_target=max_seq_len_target,
                                               batch_size=batch_size,
                                               fill_up=fill_up,
                                               use_feature_loader=use_feature_loader,
                                               preload_features=preload_features)

    return train_iter, validation_iter, config_data, data_info


class ImageTextSampleIter(ParallelSampleIter):
    """
    Data iterator on a bucketed ParallelDataSet which loads images in the source on the fly.
    It also resizes and preprocesses the images. Shuffles data at every reset and
    supports saving and loading the iterator state.
    """

    def __init__(self,
                 data: ParallelDataSet,
                 buckets,
                 batch_size,
                 bucket_batch_sizes,
                 image_size: tuple,
                 source_data_name=C.SOURCE_NAME,
                 target_data_name=C.TARGET_NAME,
                 label_name=C.TARGET_LABEL_NAME,
                 dtype='float32',
                 source_only=False,
                 use_feature_loader:bool = False,
                 preload_features: bool = False) -> None:
        super().__init__(data, buckets, batch_size, bucket_batch_sizes,
                         source_data_name, target_data_name, label_name, dtype)

        self.with_text = not source_only
        self.image_size = tuple(image_size)

        # Override provide_data to make sure to use images
        self.provide_data = [
            mx.io.DataDesc(name=self.source_data_name,
                           shape=(self.bucket_batch_sizes[-1].batch_size, ) + self.image_size,  # "NCHW"
                           layout=C.BATCH_MAJOR_IMAGE)
        ]
        if self.with_text:
            self.provide_data += [
                mx.io.DataDesc(name=self.target_data_name,
                               shape=(self.bucket_batch_sizes[-1].batch_size, self.default_bucket_key[1]),
                               layout=C.BATCH_MAJOR)
            ]
        self.use_feature_loader = use_feature_loader
        self.preload_features =preload_features
        if self.use_feature_loader:
            self.data_loader = load_features
            # Load already everything to memory
            if self.preload_features:
                logger.info("Loading all the features to memory (this might take a while, be patient)...")
                start = time.time()
                self.loaded_source = {}
                for bucket in self.data.source:
                    for k in bucket:
                        if k not in self.loaded_source:  # avoid to load twice
                            self.loaded_source[k] = load_feature(k, self.image_size)
                logger.info("Feature loaded in {} seconds.".format(time.time()-start))
        else:
            self.data_loader = load_preprocess_images

    def next(self) -> mx.io.DataBatch:
        """
        Returns the next batch from the data iterator.
        """
        if not self.iter_next():
            raise StopIteration

        i, j = self.batch_indices[self.curr_batch_index]
        self.curr_batch_index += 1

        batch_size = self.bucket_batch_sizes[i].batch_size
        source = self.data.source[i][j:j + batch_size]
        target = self.data.target[i][j:j + batch_size]
        if self.preload_features:
            loaded_source = []
            for k in source:
                loaded_source.append(self.loaded_source[k])
            loaded_source = mx.nd.array(loaded_source)
        else:
            loaded_source = mx.nd.array(self.data_loader(source, self.image_size))

        label = [self.data.label[i][j:j + batch_size]]

        provide_data = [
            mx.io.DataDesc(name=self.source_data_name, shape=loaded_source.shape, layout=C.BATCH_MAJOR_IMAGE),
        ]
        if self.with_text:
            provide_data += [
                mx.io.DataDesc(name=self.target_data_name, shape=target.shape, layout=C.BATCH_MAJOR)
        ]
        provide_label = [mx.io.DataDesc(name=n, shape=x.shape, layout=C.BATCH_MAJOR) for n, x in
                         zip(self.label_names, label)]

        data = [loaded_source]
        if self.with_text:
            data += [target]
        return mx.io.DataBatch(data, label,
                               pad=0, index=None, bucket_key=self.buckets[i],
                               provide_data=provide_data, provide_label=provide_label)

    @staticmethod
    def visualize_batch(batch: mx.io.DataBatch,
                        reverse_vocab: Dict[int, str],
                        source_only: bool =False) -> None:

        try:  # Try to import matplotlib
            import matplotlib  # pylint: disable=import-error
        except ImportError as e:
            raise RuntimeError("Please install matplotlib.")
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt

        N = M = 4
        fig, axs = plt.subplots(N, M, figsize=(20, 10))
        # Remove axes
        for i in range(N):
            for j in range(M):
                axs[i, j].axis("off")
        for i, img in enumerate(batch.data[0]):
            # (channel, height, width) -> (height, width, channel)
            img_ = np.swapaxes(img.asnumpy(), 0, 2)
            img_ = np.swapaxes(img_, 0, 1)
            axs[i//N%M, i%N].imshow(np.uint8(img_))
            axs[i//N%M, i%N].axis("off")
            if not source_only:
                sentence = ""
                sentence_ids = batch.data[1][i].asnumpy()
                carry_on = jj = 0
                for j,v in enumerate(sentence_ids):
                    if reverse_vocab[v] not in C.VOCAB_SYMBOLS:  # Ignore for visualization
                        sentence += reverse_vocab[v]
                        carry_on += len(reverse_vocab[v])
                        if jj<len(sentence_ids):
                            if carry_on>=15:
                                sentence += "\n"
                                carry_on = 0
                            else:
                                sentence += " "
                        jj += 1
                axs[i//N%M, i%N].text(0, 8, sentence, fontsize=10,
                                      bbox={'facecolor': 'white', 'alpha': 0.7, 'pad': 2})
        plt.show()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import random
import string

import pytest

from test.common_image_captioning import run_extract_features_captioning, \
    tmp_img_captioning_dataset

IMAGE_ENCODER_SETTINGS = [
    ("conv1"),
    ("conv2"),
]


@pytest.mark.parametrize("layer",
                         IMAGE_ENCODER_SETTINGS)
def test_caption_random_features(layer: str):
    source_image_size = (3, 20, 20)
    batch_size = 8
    extract_params = "--source-image-size {s1} {s2} {s3} --batch-size {batch_size} " \
                     "--image-encoder-layer {layer}".format(s1=source_image_size[0],
                                                            s2=source_image_size[1],
                                                            s3=source_image_size[2],
                                                            batch_size=batch_size,
                                                            layer=layer)

    # generate random names
    source_list = [
        ''.join(random.choice(string.ascii_uppercase) for _ in range(4)) for i
        in range(8)]
    prefix = "tmp_features"
    use_features = False
    with tmp_img_captioning_dataset(source_list,
                                    prefix,
                                    train_max_length=1,
                                    dev_max_length=1,
                                    test_max_length=1,
                                    use_features=use_features) as data:
        source_files = [data["source"], data["validation_source"],
                        data["test_source"]]
        run_extract_features_captioning(source_image_size=source_image_size,
                                        batch_size=batch_size,
                                        extract_params=extract_params,
                                        source_files=source_files,
                                        image_root=data['work_dir'])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from tempfile import TemporaryDirectory

import numpy as np
from PIL import Image

import sockeye.image_captioning.utils as utils


def test_copy_mx_model_to():
    model_path = "test"
    model_epoch = 0

    with TemporaryDirectory() as work_dir:
        # Simulate model files
        model_path = os.path.join(work_dir, model_path)
        json_name = model_path + '-symbol.json'
        params_name = model_path + '-%04d.params' % model_epoch
        open(json_name, 'a').close()
        open(params_name, 'a').close()

        with TemporaryDirectory() as output_folder:
            target_path = utils.copy_mx_model_to(model_path, model_epoch, output_folder)
            assert os.path.exists(target_path + '-symbol.json')
            assert os.path.exists(target_path + '-%04d.params' % model_epoch)


def test_crop_resize_image():
    image_size = [224, 224]
    imarray = np.random.rand(100, 250, 3) * 255
    image = Image.fromarray(imarray.astype('uint8'))
    image_o = utils.crop_resize_image(image, image_size)
    image_o = np.asarray(image_o)

    np.testing.assert_equal(image_o.shape[:2], image_size)


def test_load_preprocess_images():
    image_size = [3, 224, 224]
    image_paths = ['a.jpg', 'b.jpg', 'c.jpg']
    # Generate a set of images
    with TemporaryDirectory() as work_dir:
        filenames = []
        for s in image_paths:
            filename = os.path.join(work_dir, s)
            imarray = np.random.rand(100, 100, 3) * 255
            im = Image.fromarray(imarray.astype('uint8'))
            im.save(filename)
            filenames.append(filename)

        images = utils.load_preprocess_images(filenames, image_size)
        assert len(images)==3
        for img in images:
            np.testing.assert_equal(img.shape, image_size)


def test_load_features():
    feature_size = [10, 2048]
    filenames = ['a.npy', 'b.npy', 'c.npy', 'd.npy']
    # Generate a set of images
    with TemporaryDirectory() as work_dir:
        paths = []
        for s in filenames:
            filename = os.path.join(work_dir, s)
            data = np.random.rand(*feature_size)
            np.save(filename, data)
            paths.append(filename)

        feats = utils.load_features(paths, feature_size)
        assert len(feats)==4
        for f in feats:
            np.testing.assert_equal(f.shape, feature_size)


def test_save_features():
    feature_size = [10, 2048]
    filenames = ['a', 'b', 'c']
    # Generate the list of ndarrays
    datas = []
    for i in range(len(filenames)):
        datas.append(np.random.rand(*feature_size))

    with TemporaryDirectory() as work_dir:
        paths = [os.path.join(work_dir, s) for s in filenames]
        fnames = utils.save_features(paths, datas)
        for i, f in enumerate(fnames):
            assert os.path.exists(f)>0
            data = utils.load_feature(f, feature_size)
            np.testing.assert_almost_equal(datas[i], data)

    # Tests with compression
    with TemporaryDirectory() as work_dir:
        paths = [os.path.join(work_dir, s) for s in filenames]
        fnames = utils.save_features(paths, datas, compressed=True)
        for i, f in enumerate(fnames):
            assert os.path.exists(f)>0
            data = utils.load_feature(f, feature_size)
            np.testing.assert_almost_equal(datas[i], data)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You may not
# use this file except in compliance with the License. A copy of the License
# is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is distributed on
# an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import argparse
import pytest

import sockeye.image_captioning.arguments as arguments
import sockeye.constants as C

from test.unit.test_arguments import _test_args


@pytest.mark.parametrize("test_params, expected_params", [
    ('--image-root test_img_root --input input --output-root test_output_root --output output',
     dict(source_image_size=[3, 224, 224],
          image_root="test_img_root",
          input="input",
          output_root="test_output_root",
          output="output",
          batch_size=64,
          image_positional_embedding_type=C.NO_POSITIONAL_EMBEDDING,
          image_encoder_model_path="/path/to/mxnet/image/model/",
          image_encoder_model_epoch=0,
          image_encoder_layer="stage4_unit3_conv3",
          image_encoder_conv_map_size=49,
          image_encoder_num_hidden=512,
          no_image_encoder_global_descriptor=True,
          load_all_features_to_memory=False,
          device_ids=[-1],
          disable_device_locking=False,
          lock_dir='/tmp',
          use_cpu=False,
          extract_image_features=False
     ))
])
def test_image_extract_features_cli_args(test_params, expected_params):
    _test_args(test_params, expected_params, arguments.add_image_extract_features_cli_args)


@pytest.mark.parametrize("test_params, expected_params", [
    ('--source-root test_src_root',
     dict(source_root="test_src_root"))
])
def test_image_source_root_args(test_params, expected_params):
    _test_args(test_params, expected_params, arguments.add_image_source_root_args)


@pytest.mark.parametrize("test_params, expected_params", [
    ('--validation-source-root test_val_src_root --validation-source val_src --validation-target val_tgt',
     dict(validation_source_root="test_val_src_root",
          validation_source="val_src",
          validation_target="val_tgt",
          validation_source_factors=[]
     ))
])
def test_image_validation_data_params(test_params, expected_params):
    _test_args(test_params, expected_params, arguments.add_image_validation_data_params)


@pytest.mark.parametrize("test_params, expected_params", [
    ('--load-all-features-to-memory',
     dict(load_all_features_to_memory=True, extract_image_features=False))
])
def test_preextracted_features_args(test_params, expected_params):
    _test_args(test_params, expected_params, arguments.add_preextracted_features_args)


def test_add_image_train_cli_args():
     # Just make sure that it does not fail. We covered above the main tests and
     # the rest are coveder in test/unit/test_arguments.py
     params = argparse.ArgumentParser()
     arguments.add_image_train_cli_args(params)


def test_add_image_caption_cli_args():
     # Just make sure that it does not fail. We covered above the main tests and
     # the rest are coveder in test/unit/test_arguments.py
     params = argparse.ArgumentParser()
     arguments.add_image_caption_cli_args(params)
